#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import breezy

from breezy.tests import (
    TestCase,
)


from breezy.bzr import RemoteBzrProber
from breezy.git import RemoteGitProber

from ..debian import (
    select_probers,
    convert_debian_vcs_url,
    UnsupportedVCSProber,
)


class SelectProbersTests(TestCase):
    def test_none(self):
        self.assertIs(None, select_probers())
        self.assertIs(None, select_probers(None))

    def test_bzr(self):
        self.assertEqual([RemoteBzrProber], select_probers("bzr"))

    def test_git(self):
        self.assertEqual([RemoteGitProber], select_probers("git"))

    def test_unsupported(self):
        self.assertEqual([UnsupportedVCSProber("foo")], select_probers("foo"))


class ConvertDebianVcsUrlTests(TestCase):
    def test_git(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/blah.git",
            convert_debian_vcs_url("Git", "https://salsa.debian.org/jelmer/blah.git"),
        )

    def test_git_ssh(self):
        if breezy.version_info < (3, 1, 1):
            self.knownFailure("breezy < 3.1.1 can not deal with ssh:// URLs")
        self.assertEqual(
            "ssh://git@git.kali.org/jelmer/blah.git",
            convert_debian_vcs_url("Git", "ssh://git@git.kali.org/jelmer/blah.git"),
        )



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from debian.deb822 import Deb822
from debian.changelog import Version
import os
import subprocess
from typing import Optional, Dict, List

from debmutate.vcs import split_vcs_url

from breezy import urlutils
from breezy.branch import Branch
from breezy.errors import UnsupportedFormatError
from breezy.controldir import Prober, ControlDirFormat
from breezy.bzr import RemoteBzrProber
from breezy.git import RemoteGitProber
from breezy.git.repository import GitRepository
from breezy.plugins.debian.cmds import cmd_builddeb
from breezy.plugins.debian.directory import (
    source_package_vcs,
    vcs_field_to_bzr_url_converters,
)

from breezy.tree import Tree
from breezy.urlutils import InvalidURL
from breezy.workingtree import WorkingTree

from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.upstream import (
    MissingUpstreamTarball,
)

from lintian_brush.detect_gbp_dch import guess_update_changelog
from lintian_brush.changelog import add_changelog_entry

from .. import proposal as _mod_proposal
from ..utils import (
    open_branch,
)


__all__ = [
    "add_changelog_entry",
    "changelog_add_line",
    "apt_get_source_package",
    "guess_update_changelog",
    "source_package_vcs",
    "build",
    "BuildFailedError",
    "MissingUpstreamTarball",
    "vcs_field_to_bzr_url_converters",
]


DEFAULT_BUILDER = "sbuild --no-clean-source"


class NoSuchPackage(Exception):
    """No such package."""


def build(
    tree: Tree,
    subpath: str = "",
    builder: Optional[str] = None,
    result_dir: Optional[str] = None,
) -> None:
    """Build a debian package in a directory.

    Args:
      tree: Working tree
      subpath: Subpath to build in
      builder: Builder command (e.g. 'sbuild', 'debuild')
      result_dir: Directory to copy results to
    """
    if builder is None:
        builder = DEFAULT_BUILDER
    # TODO(jelmer): Refactor brz-debian so it's not necessary
    # to call out to cmd_builddeb, but to lower-level
    # functions instead.
    cmd_builddeb().run([tree.abspath(subpath)], builder=builder, result_dir=result_dir)


class NoAptSources(Exception):
    """No apt sources were configured."""


def apt_get_source_package(name: str) -> Deb822:
    """Get source package metadata.

    Args:
      name: Name of the source package
    Returns:
      A `Deb822` object
    """
    import apt_pkg

    apt_pkg.init()

    try:
        sources = apt_pkg.SourceRecords()
    except apt_pkg.Error as e:
        if e.args[0] == ("E:You must put some 'deb-src' URIs in your sources.list"):
            raise NoAptSources()
        raise

    by_version: Dict[str, Deb822] = {}
    while sources.lookup(name):
        by_version[sources.version] = sources.record  # type: ignore

    if len(by_version) == 0:
        raise NoSuchPackage(name)

    # Try the latest version
    version = sorted(by_version, key=Version)[-1]

    return Deb822(by_version[version])


def convert_debian_vcs_url(vcs_type: str, vcs_url: str) -> str:
    converters = dict(vcs_field_to_bzr_url_converters)
    try:
        return converters[vcs_type](vcs_url)
    except KeyError:
        raise ValueError("unknown vcs %s" % vcs_type)
    except InvalidURL as e:
        raise ValueError("invalid URL: %s" % e)


def open_packaging_branch(location, possible_transports=None, vcs_type=None):
    """Open a packaging branch from a location string.

    location can either be a package name or a full URL
    """
    if "/" not in location and ":" not in location:
        pkg_source = apt_get_source_package(location)
        try:
            (vcs_type, vcs_url) = source_package_vcs(pkg_source)
        except KeyError:
            raise Exception("Package %s does not have any VCS information" % location)
        (url, branch_name, subpath) = split_vcs_url(vcs_url)
    else:
        url, params = urlutils.split_segment_parameters(location)
        branch_name = params.get("branch")
        subpath = ""
    probers = select_probers(vcs_type)
    branch = open_branch(
        url, possible_transports=possible_transports, probers=probers, name=branch_name
    )
    return branch, subpath or ""


def pick_additional_colocated_branches(main_branch):
    ret = ["pristine-tar", "pristine-lfs", "upstream"]
    ret.append("patch-queue/" + main_branch.name)
    if main_branch.name.startswith("debian/"):
        parts = main_branch.name.split("/")
        parts[0] = "upstream"
        ret.append("/".join(parts))
    return ret


class Workspace(_mod_proposal.Workspace):
    def __init__(self, main_branch: Branch, *args, **kwargs) -> None:
        if isinstance(main_branch.repository, GitRepository):
            kwargs["additional_colocated_branches"] = kwargs.get(
                "additional_colocated_branches", []
            ) + pick_additional_colocated_branches(main_branch)
        super(Workspace, self).__init__(main_branch, *args, **kwargs)

    def build(
        self,
        builder: Optional[str] = None,
        result_dir: Optional[str] = None,
        subpath: str = "",
    ) -> None:
        return build(
            tree=self.local_tree,
            subpath=subpath,
            builder=builder,
            result_dir=result_dir,
        )


class UnsupportedVCSProber(Prober):
    def __init__(self, vcs_type):
        self.vcs_type = vcs_type

    def __eq__(self, other):
        return isinstance(other, type(self)) and other.vcs_type == self.vcs_type

    def __call__(self):
        # The prober expects to be registered as a class.
        return self

    def priority(self, transport):
        return 200

    def probe_transport(self, transport):
        raise UnsupportedFormatError(
            "This VCS %s is not currently supported." % self.vcs_type
        )

    @classmethod
    def known_formats(klass):
        return []


prober_registry = {
    "bzr": RemoteBzrProber,
    "git": RemoteGitProber,
}

try:
    from breezy.plugins.fossil import RemoteFossilProber
except ImportError:
    pass
else:
    prober_registry["fossil"] = RemoteFossilProber

try:
    from breezy.plugins.svn import SvnRepositoryProber
except ImportError:
    pass
else:
    prober_registry["svn"] = SvnRepositoryProber

try:
    from breezy.plugins.hg import SmartHgProber
except ImportError:
    pass
else:
    prober_registry["hg"] = SmartHgProber

try:
    from breezy.plugins.darcs import DarcsProber
except ImportError:
    pass
else:
    prober_registry["darcs"] = DarcsProber

try:
    from breezy.plugins.cvs import CVSProber
except ImportError:
    pass
else:
    prober_registry["cvs"] = CVSProber


def select_probers(vcs_type=None):
    if vcs_type is None:
        return None
    try:
        return [prober_registry[vcs_type.lower()]]
    except KeyError:
        return [UnsupportedVCSProber(vcs_type)]


def select_preferred_probers(vcs_type: Optional[str] = None) -> List[Prober]:
    probers = list(ControlDirFormat.all_probers())
    if vcs_type:
        try:
            probers.insert(0, prober_registry[vcs_type.lower()])
        except KeyError:
            pass
    return probers


def changelog_add_line(
    tree: WorkingTree, subpath: str, line: str, email: Optional[str] = None
) -> None:
    env = {}
    if email:
        env["DEBEMAIL"] = email
    subprocess.check_call(["dch", "--", line], cwd=tree.abspath(subpath), env=env)


def is_debcargo_package(tree: Tree, subpath: str) -> bool:
    debian_path = os.path.join(subpath, "debian")
    if tree.has_filename(debian_path):
        return False
    control_path = os.path.join(subpath, "debcargo.toml")
    if tree.has_filename(control_path):
        return True
    return False


def control_files_in_root(tree: Tree, subpath: str) -> bool:
    debian_path = os.path.join(subpath, "debian")
    if tree.has_filename(debian_path):
        return False
    control_path = os.path.join(subpath, "control")
    if tree.has_filename(control_path):
        return True
    if tree.has_filename(control_path + ".in"):
        return True
    return False


def control_file_present(tree: Tree, subpath: str) -> bool:
    """Check whether there are any control files present in a tree.

    Args:
      tree: Tree to check
      subpath: subpath to check
    Returns:
      whether control file is present
    """
    for name in ["debian/control", "debian/control.in", "control", "control.in"]:
        name = os.path.join(subpath, name)
        if tree.has_filename(name):
            return True
    return False



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
from urllib.parse import urlparse

from breezy import osutils

from debmutate.control import ControlEditor
from debmutate.reformatting import GeneratedFile, FormattingUnpreservable


from . import (
    pick_additional_colocated_branches,
    add_changelog_entry,
)
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerError,
    ChangerResult,
)
from ..proposal import push_changes


BRANCH_NAME = "orphan"


def push_to_salsa(local_tree, user, name, dry_run=False):
    from breezy.branch import Branch
    from breezy.plugins.gitlab.hoster import GitLab

    salsa = GitLab.probe_from_url("https://salsa.debian.org/")
    # TODO(jelmer): Fork if the old branch was hosted on salsa
    if dry_run:
        logging.info("Creating and pushing to salsa project %s/%s", user, name)
        return
    salsa.create_project("%s/%s" % (user, name))
    target_branch = Branch.open(
        "git+ssh://git@salsa.debian.org/%s/%s.git" % (user, name)
    )
    additional_colocated_branches = pick_additional_colocated_branches(
        local_tree.branch
    )
    return push_changes(
        local_tree.branch,
        target_branch,
        hoster=salsa,
        additional_colocated_branches=additional_colocated_branches,
        dry_run=dry_run,
    )


class OrphanResult(object):
    def __init__(
        self, package=None, old_vcs_url=None, new_vcs_url=None, salsa_user=None
    ):
        self.package = package
        self.old_vcs_url = old_vcs_url
        self.new_vcs_url = new_vcs_url
        self.pushed = False
        self.salsa_user = salsa_user


class OrphanChanger(DebianChanger):

    name = "orphan"

    def __init__(
        self, update_vcs=True, salsa_push=True, salsa_user="debian", dry_run=False
    ):
        self.update_vcs = update_vcs
        self.salsa_push = salsa_push
        self.salsa_user = salsa_user
        self.dry_run = dry_run

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument(
            "--no-update-vcs",
            action="store_true",
            help="Do not move the VCS repository to the Debian team on Salsa.",
        )
        parser.add_argument(
            "--salsa-user",
            type=str,
            default="debian",
            help="Salsa user to push repository to.",
        )
        parser.add_argument(
            "--just-update-headers",
            action="store_true",
            help="Update the VCS-* headers, but don't actually "
            "clone the repository.",
        )
        parser.add_argument("--dry-run", action="store_true", help="Dry run changes.")

    @classmethod
    def from_args(cls, args):
        return cls(
            update_vcs=not args.no_update_vcs,
            dry_run=args.dry_run,
            salsa_user=args.salsa_user,
            salsa_push=not args.just_update_headers,
        )

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):
        base_revid = local_tree.last_revision()
        control_path = local_tree.abspath(osutils.pathjoin(subpath, "debian/control"))
        try:
            with ControlEditor(path=control_path) as editor:
                editor.source["Maintainer"] = "Debian QA Group <packages@qa.debian.org>"
                try:
                    del editor.source["Uploaders"]
                except KeyError:
                    pass

            result = OrphanResult()

            if self.update_vcs:
                with ControlEditor(path=control_path) as editor:
                    result.package_name = editor.source["Source"]
                    result.old_vcs_url = editor.source.get("Vcs-Git")
                    editor.source["Vcs-Git"] = "https://salsa.debian.org/%s/%s.git" % (
                        self.salsa_user,
                        result.package_name,
                    )
                    result.new_vcs_url = editor.source["Vcs-Git"]
                    editor.source["Vcs-Browser"] = "https://salsa.debian.org/%s/%s" % (
                        self.salsa_user,
                        result.package_name,
                    )
                    result.salsa_user = self.salsa_user
                if result.old_vcs_url == result.new_vcs_url:
                    result.old_vcs_url = result.new_vcs_url = None
            if update_changelog in (True, None):
                add_changelog_entry(
                    local_tree,
                    osutils.pathjoin(subpath, "debian/changelog"),
                    ["QA Upload.", "Move package to QA team."],
                )
            local_tree.commit(
                "Move package to QA team.", committer=committer, allow_pointless=False
            )
        except FormattingUnpreservable as e:
            raise ChangerError(
                "formatting-unpreservable",
                "unable to preserve formatting while editing %s" % e.path,
            )
        except GeneratedFile as e:
            raise ChangerError(
                "generated-file", "unable to edit generated file: %r" % e
            )

        if self.update_vcs and self.salsa_push and result.new_vcs_url:
            push_to_salsa(
                local_tree, self.salsa_user, result.package_name, dry_run=self.dry_run
            )
            result.pushed = True
            reporter.report_metadata("old_vcs_url", result.old_vcs_url)
            reporter.report_metadata("new_vcs_url", result.new_vcs_url)
            reporter.report_metadata("pushed", result.pushed)

        branches = [("main", None, base_revid, local_tree.last_revision())]

        tags = []

        return ChangerResult(
            description="Move package to QA team.",
            mutator=result,
            branches=branches,
            tags=tags,
            sufficient_for_proposal=True,
            proposed_commit_message=("Set the package maintainer to the QA team."),
        )

    def get_proposal_description(self, applied, description_format, existing_proposal):
        return "Set the package maintainer to the QA team."

    def describe(self, result, publish_result):
        if publish_result.is_new:
            logging.info(
                "Proposed change of maintainer to QA team: %s",
                publish_result.proposal.url,
            )
        else:
            logging.info(
                "No changes for orphaned package %s",
                result.package_name)
        if result.pushed:
            logging.info("Pushed new package to %s.", result.new_vcs_url)
        elif result.new_vcs_url:
            for line in move_instructions(
                result.package_name,
                result.salsa_user,
                result.old_vcs_url,
                result.new_vcs_url,
            ):
                logging.info("%s", line)

    @classmethod
    def describe_command(cls, command):
        return "Mark as orphaned"


def move_instructions(package_name, salsa_user, old_vcs_url, new_vcs_url):
    yield "Please move the repository from %s to %s." % (old_vcs_url, new_vcs_url)
    if urlparse(old_vcs_url).hostname == "salsa.debian.org":
        path = urlparse(old_vcs_url).path
        if path.endswith(".git"):
            path = path[:-4]
        yield "If you have the salsa(1) tool installed, run: "
        yield ""
        yield "    salsa fork --group=%s %s" % (salsa_user, path)
    else:
        yield "If you have the salsa(1) tool installed, run: "
        yield ""
        yield "    git clone %s %s" % (old_vcs_url, package_name)
        yield "    salsa --group=%s push_repo %s" % (salsa_user, package_name)


if __name__ == "__main__":
    import sys

    sys.exit(run_mutator(OrphanChanger))



# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import logging
import sys

import breezy
from breezy.revision import NULL_REVISION

from lintian_brush import (
    version_string as lintian_brush_version_string,
)
from lintian_brush.debianize import (
    debianize,
)
from lintian_brush.config import Config

import silver_platter

from .changer import (
    DebianChanger,
    run_mutator,
    ChangerResult,
)


BRANCH_NAME = "debianize"


class DebianizeChanger(DebianChanger):

    name = "debianize"

    def __init__(self, compat_release=None):
        self.compat_release = compat_release

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument("--compat-release", type=str, help=argparse.SUPPRESS)

    @classmethod
    def from_args(cls, args):
        return cls(compat_release=args.compat_release)

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):
        base_revid = local_tree.last_revision()
        upstream_base_revid = NULL_REVISION

        reporter.report_metadata(
            "versions",
            {
                "lintian-brush": lintian_brush_version_string,
                "silver-platter": silver_platter.version_string,
                "breezy": breezy.version_string,
            },
        )

        import distro_info

        debian_info = distro_info.DebianDistroInfo()

        compat_release = self.compat_release
        try:
            cfg = Config.from_workingtree(local_tree, subpath)
        except FileNotFoundError:
            pass
        else:
            compat_release = cfg.compat_release()
            if compat_release:
                compat_release = debian_info.codename(
                    compat_release, default=compat_release
                )
        if compat_release is None:
            compat_release = debian_info.stable()

        with local_tree.lock_write():
            debianize(local_tree, subpath=subpath, compat_release=self.compat_release)

        # TODO(jelmer): Pristine tar branch?
        branches = [
            ("main", None, base_revid, local_tree.last_revision()),
            (
                "upstream",
                "upstream",
                upstream_base_revid,
                local_tree.controldir.open_branch("upstream").last_revision(),
            ),
        ]

        return ChangerResult(
            description="Debianized package.",
            mutator=None,
            branches=branches,
            tags=[],
            value=None,
            sufficient_for_proposal=True,
        )

    def get_proposal_description(self, applied, description_format, existing_proposal):
        return "Debianize package."

    def describe(self, applied, publish_result):
        logging.info("Created Debian package.")


if __name__ == "__main__":
    sys.exit(run_mutator(DebianizeChanger))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
from typing import List, Union, Dict, Optional, Tuple, Any, Callable

from breezy.branch import Branch
from breezy import (
    errors,
    merge as _mod_merge,
    revision as _mod_revision,
)
from breezy.errors import PermissionDenied
from breezy.propose import (
    get_hoster,
    Hoster,
    MergeProposal,
    MergeProposalExists,
    NoSuchProject,
    UnsupportedHoster,
)
from breezy.transport import Transport

try:
    from breezy.propose import (
        SourceNotDerivedFromTarget,
    )
except ImportError:  # breezy < 3.1.1
    SourceNotDerivedFromTarget = None


from .utils import (
    open_branch,
    MemoryBranch,
    full_branch_url,
)


__all__ = [
    "push_changes",
    "push_derived_changes",
    "propose_changes",
    "EmptyMergeProposal",
    "check_proposal_diff",
    "DryRunProposal",
    "find_existing_proposed",
    "NoSuchProject",
    "PermissionDenied",
    "UnsupportedHoster",
]


MODE_PUSH = "push"
MODE_ATTEMPT_PUSH = "attempt-push"
MODE_PROPOSE = "propose"
MODE_PUSH_DERIVED = "push-derived"
SUPPORTED_MODES: List[str] = [
    MODE_PUSH,
    MODE_ATTEMPT_PUSH,
    MODE_PROPOSE,
    MODE_PUSH_DERIVED,
]


def _tag_selector_from_tags(tags):
    # TODO(jelmer): Select dict
    return tags.__contains__


def push_result(
    local_branch: Branch,
    remote_branch: Branch,
    additional_colocated_branches: Optional[List[str]] = None,
    tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
    stop_revision: Optional[bytes] = None,
) -> None:
    kwargs = {}
    if tags is not None:
        kwargs["tag_selector"] = _tag_selector_from_tags(tags)
    try:
        local_branch.push(
            remote_branch, overwrite=False, stop_revision=stop_revision, **kwargs
        )
    except errors.LockFailed as e:
        # Almost certainly actually a PermissionDenied error..
        raise errors.PermissionDenied(path=full_branch_url(remote_branch), extra=e)
    for branch_name in additional_colocated_branches or []:
        try:
            add_branch = local_branch.controldir.open_branch(name=branch_name)
        except errors.NotBranchError:
            pass
        else:
            remote_branch.controldir.push_branch(add_branch, name=branch_name, **kwargs)


def push_changes(
    local_branch: Branch,
    main_branch: Branch,
    hoster: Optional[Hoster],
    possible_transports: Optional[List[Transport]] = None,
    additional_colocated_branches: Optional[List[str]] = None,
    dry_run: bool = False,
    tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
    stop_revision: Optional[bytes] = None,
) -> None:
    """Push changes to a branch."""
    if hoster is None:
        push_url = main_branch.user_url
    else:
        push_url = hoster.get_push_url(main_branch)
    logging.info("pushing to %s", push_url)
    target_branch = open_branch(push_url, possible_transports=possible_transports)
    if not dry_run:
        push_result(
            local_branch,
            target_branch,
            additional_colocated_branches,
            tags=tags,
            stop_revision=stop_revision,
        )


def push_derived_changes(
    local_branch: Branch,
    main_branch: Branch,
    hoster: Hoster,
    name: str,
    overwrite_existing: Optional[bool] = False,
    owner: Optional[str] = None,
    tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
    stop_revision: Optional[bytes] = None,
) -> Tuple[Branch, str]:
    kwargs = {}
    if tags is not None:
        kwargs["tag_selector"] = _tag_selector_from_tags(tags)
    remote_branch, public_branch_url = hoster.publish_derived(
        local_branch,
        main_branch,
        name=name,
        overwrite=overwrite_existing,
        owner=owner,
        revision_id=stop_revision,
        **kwargs
    )
    return remote_branch, public_branch_url


def propose_changes(  # noqa: C901
    local_branch: Branch,
    main_branch: Branch,
    hoster: Hoster,
    name: str,
    mp_description: str,
    resume_branch: Optional[Branch] = None,
    resume_proposal: Optional[MergeProposal] = None,
    overwrite_existing: Optional[bool] = True,
    labels: Optional[List[str]] = None,
    dry_run: bool = False,
    commit_message: Optional[str] = None,
    additional_colocated_branches: Optional[List[str]] = None,
    allow_empty: bool = False,
    reviewers: Optional[List[str]] = None,
    tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
    owner: Optional[str] = None,
    stop_revision: Optional[bytes] = None,
    allow_collaboration: bool = False,
) -> Tuple[MergeProposal, bool]:
    """Create or update a merge proposal.

    Args:
      local_branch: Local branch with changes to propose
      main_branch: Target branch to propose against
      hoster: Associated hoster for main branch
      mp_description: Merge proposal description
      resume_branch: Existing derived branch
      resume_proposal: Existing merge proposal to resume
      overwrite_existing: Whether to overwrite any other existing branch
      labels: Labels to add
      dry_run: Whether to just dry-run the change
      commit_message: Optional commit message
      additional_colocated_branches: Additional colocated branches to propose
      allow_empty: Whether to allow empty merge proposals
      reviewers: List of reviewers
      tags: Tags to push (None for default behaviour)
      owner: Derived branch owner
      stop_revision: Revision to stop pushing at
      allow_collaboration: Allow target branch owners to modify source branch
    Returns:
      Tuple with (proposal, is_new)
    """
    if not allow_empty:
        check_proposal_diff(local_branch, main_branch, stop_revision)
    push_kwargs = {}
    if tags is not None:
        push_kwargs["tag_selector"] = _tag_selector_from_tags(tags)
    if not dry_run:
        if resume_branch is not None:
            local_branch.push(
                resume_branch,
                overwrite=overwrite_existing,
                stop_revision=stop_revision,
                **push_kwargs
            )
            remote_branch = resume_branch
        else:
            remote_branch, public_branch_url = hoster.publish_derived(
                local_branch,
                main_branch,
                name=name,
                overwrite=overwrite_existing,
                revision_id=stop_revision,
                owner=owner,
                **push_kwargs
            )
        for colocated_branch_name in additional_colocated_branches or []:
            try:
                local_colo_branch = local_branch.controldir.open_branch(
                    name=colocated_branch_name
                )
            except errors.NotBranchError:
                pass
            else:
                remote_branch.controldir.push_branch(
                    source=local_colo_branch,
                    overwrite=overwrite_existing,
                    name=colocated_branch_name,
                    **push_kwargs
                )
    if resume_proposal is not None and dry_run:
        resume_proposal = DryRunProposal.from_existing(
            resume_proposal, source_branch=local_branch
        )
    if (
        resume_proposal is not None
        and getattr(resume_proposal, "is_closed", None)
        and resume_proposal.is_closed()
    ):
        from breezy.propose import (
            ReopenFailed,
        )

        try:
            resume_proposal.reopen()
        except ReopenFailed:
            logging.info(
                "Reopening existing proposal failed. Creating new proposal.")
            resume_proposal = None
    if resume_proposal is None:
        if not dry_run:
            proposal_builder = hoster.get_proposer(remote_branch, main_branch)
            kwargs: Dict[str, Any] = {}
            kwargs["commit_message"] = commit_message
            kwargs["allow_collaboration"] = allow_collaboration
            try:
                mp = proposal_builder.create_proposal(
                    description=mp_description,
                    labels=labels,
                    reviewers=reviewers,
                    **kwargs
                )
            except MergeProposalExists as e:
                if getattr(e, "existing_proposal", None) is None:
                    # Hoster didn't tell us where the actual proposal is.
                    raise
                resume_proposal = e.existing_proposal
            except errors.PermissionDenied:
                logging.info(
                    "Permission denied while trying to create " "proposal.")
                raise
            else:
                return (mp, True)
        else:
            mp = DryRunProposal(
                local_branch,
                main_branch,
                labels=labels,
                description=mp_description,
                commit_message=commit_message,
                reviewers=reviewers,
                owner=owner,
                stop_revision=stop_revision,
            )
            return (mp, True)
    # Check that the proposal doesn't already has this description.
    # Setting the description (regardless of whether it changes)
    # causes Launchpad to send emails.
    if resume_proposal.get_description() != mp_description:
        resume_proposal.set_description(mp_description)
    if resume_proposal.get_commit_message() != commit_message:
        try:
            resume_proposal.set_commit_message(commit_message)
        except errors.UnsupportedOperation:
            pass
    return (resume_proposal, False)


class EmptyMergeProposal(Exception):
    """Merge proposal does not have any changes."""

    def __init__(self, local_branch: Branch, main_branch: Branch):
        self.local_branch = local_branch
        self.main_branch = main_branch


def check_proposal_diff(
    other_branch: Branch, main_branch: Branch, stop_revision: Optional[bytes] = None
) -> None:
    if stop_revision is None:
        stop_revision = other_branch.last_revision()
    main_revid = main_branch.last_revision()
    other_branch.repository.fetch(main_branch.repository, main_revid)
    with other_branch.lock_read():
        main_tree = other_branch.repository.revision_tree(main_revid)
        revision_graph = other_branch.repository.get_graph()
        tree_branch = MemoryBranch(other_branch.repository, (None, main_revid), None)
        merger = _mod_merge.Merger(
            tree_branch, this_tree=main_tree, revision_graph=revision_graph
        )
        merger.set_other_revision(stop_revision, other_branch)
        try:
            merger.find_base()
        except errors.UnrelatedBranches:
            merger.set_base_revision(_mod_revision.NULL_REVISION, other_branch)
        merger.merge_type = _mod_merge.Merge3Merger
        tree_merger = merger.make_merger()
        with tree_merger.make_preview_transform() as tt:
            changes = tt.iter_changes()
            if not any(changes):
                raise EmptyMergeProposal(other_branch, main_branch)


class DryRunProposal(MergeProposal):
    """A merge proposal that is not actually created.

    :ivar url: URL for the merge proposal
    """

    def __init__(
        self,
        source_branch: Branch,
        target_branch: Branch,
        labels: Optional[List[str]] = None,
        description: Optional[str] = None,
        commit_message: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
        owner: Optional[str] = None,
        stop_revision: Optional[bytes] = None,
    ):
        self.description = description
        self.closed = False
        self.labels = labels or []
        self.source_branch = source_branch
        self.target_branch = target_branch
        self.commit_message = commit_message
        self.url = None
        self.reviewers = reviewers
        self.owner = owner
        self.stop_revision = stop_revision

    @classmethod
    def from_existing(
        cls, mp: MergeProposal, source_branch: Optional[Branch] = None
    ) -> MergeProposal:
        if source_branch is None:
            source_branch = open_branch(mp.get_source_branch_url())
        commit_message = mp.get_commit_message()
        return cls(
            source_branch=source_branch,
            target_branch=open_branch(mp.get_target_branch_url()),
            description=mp.get_description(),
            commit_message=commit_message,
        )

    def __repr__(self) -> str:
        return "%s(%r, %r)" % (
            self.__class__.__name__,
            self.source_branch,
            self.target_branch,
        )

    def get_description(self) -> Optional[str]:
        """Get the description of the merge proposal."""
        return self.description

    def set_description(self, description: str) -> None:
        self.description = description

    def get_commit_message(self) -> Optional[str]:
        return self.commit_message

    def set_commit_message(self, commit_message: str) -> None:
        self.commit_message = commit_message

    def get_source_branch_url(self) -> str:
        """Return the source branch."""
        return full_branch_url(self.source_branch)

    def get_target_branch_url(self) -> str:
        """Return the target branch."""
        return full_branch_url(self.target_branch)

    def close(self) -> None:
        """Close the merge proposal (without merging it)."""
        self.closed = True

    def is_merged(self) -> bool:
        """Check whether this merge proposal has been merged."""
        return False

    def is_closed(self) -> bool:
        """Check whether this merge proposal has been closed."""
        return False

    def reopen(self) -> None:
        pass


def find_existing_proposed(
    main_branch: Branch,
    hoster: Hoster,
    name: str,
    overwrite_unrelated: bool = False,
    owner: Optional[str] = None,
) -> Tuple[Optional[Branch], Optional[bool], Optional[MergeProposal]]:
    """Find an existing derived branch with the specified name, and proposal.

    Args:
      main_branch: Main branch
      hoster: The hoster
      name: Name of the derived branch
      overwrite_unrelated: Whether to overwrite existing (but unrelated)
        branches
    Returns:
      Tuple with (resume_branch, overwrite_existing, existing_proposal)
      The resume_branch is the branch to continue from; overwrite_existing
      means there is an existing branch in place that should be overwritten.
    """
    try:
        existing_branch = hoster.get_derived_branch(main_branch, name=name, owner=owner)
    except errors.NotBranchError:
        return (None, None, None)
    else:
        logging.info(
            "Branch %s already exists (branch at %s)",
            name,
            full_branch_url(existing_branch),
        )
        # If there is an open or rejected merge proposal, resume that.
        merged_proposal = None
        for mp in hoster.iter_proposals(existing_branch, main_branch, status="all"):
            if not mp.is_closed() and not mp.is_merged():
                return (existing_branch, False, mp)
            else:
                merged_proposal = mp
        else:
            if merged_proposal is not None:
                logging.info(
                    "There is a proposal that has already been merged at %s.",
                    merged_proposal.url,
                )
                return (None, True, None)
            else:
                # No related merge proposals found, but there is an existing
                # branch (perhaps for a different target branch?)
                if overwrite_unrelated:
                    return (None, True, None)
                else:
                    # TODO(jelmer): What to do in this case?
                    return (None, False, None)


def merge_conflicts(
    main_branch: Branch, other_branch: Branch, other_revision: Optional[bytes] = None
) -> bool:
    """Check whether two branches are conflicted when merged.

    Args:
      main_branch: Main branch to merge into
      other_branch: Branch to merge (and use for scratch access, needs write
                    access)
      other_revision: Other revision to check
    Returns:
      boolean indicating whether the merge would result in conflicts
    """
    if other_revision is None:
        other_revision = other_branch.last_revision()
    if other_branch.repository.get_graph().is_ancestor(
        main_branch.last_revision(), other_revision
    ):
        return False

    other_branch.repository.fetch(
        main_branch.repository, revision_id=main_branch.last_revision()
    )

    # Reset custom merge hooks, since they could make it harder to detect
    # conflicted merges that would appear on the hosting site.
    old_file_content_mergers = _mod_merge.Merger.hooks["merge_file_content"]
    _mod_merge.Merger.hooks["merge_file_content"] = []

    other_tree = other_branch.repository.revision_tree(other_revision)
    try:
        try:
            merger = _mod_merge.Merger.from_revision_ids(
                other_tree,
                other_branch=other_branch,
                other=main_branch.last_revision(),
                tree_branch=other_branch,
            )
        except errors.UnrelatedBranches:
            # Unrelated branches don't technically *have* to lead to
            # conflicts, but there's not a lot to be salvaged here, either.
            return True
        merger.merge_type = _mod_merge.Merge3Merger
        tree_merger = merger.make_merger()
        with tree_merger.make_preview_transform():
            return bool(tree_merger.cooked_conflicts)
    finally:
        _mod_merge.Merger.hooks["merge_file_content"] = old_file_content_mergers


class PublishResult(object):
    """A object describing the result of a publish action."""

    def __init__(
        self, mode: str, proposal: Optional[MergeProposal] = None, is_new: bool = False
    ) -> None:
        self.mode = mode
        self.proposal = proposal
        self.is_new = is_new

    def __tuple__(self) -> Tuple[Optional[MergeProposal], bool]:
        # Backwards compatibility
        return (self.proposal, self.is_new)


def publish_changes(
    local_branch: Branch,
    main_branch: Branch,
    resume_branch: Optional[Branch],
    mode: str,
    name: str,
    get_proposal_description: Callable[[str, Optional[MergeProposal]], str],
    get_proposal_commit_message: Callable[
        [Optional[MergeProposal]], Optional[str]
    ] = None,
    dry_run: bool = False,
    hoster: Optional[Hoster] = None,
    allow_create_proposal: bool = True,
    labels: Optional[List[str]] = None,
    overwrite_existing: Optional[bool] = True,
    existing_proposal: Optional[MergeProposal] = None,
    reviewers: Optional[List[str]] = None,
    tags: Optional[Union[List[str], Dict[str, bytes]]] = None,
    derived_owner: Optional[str] = None,
    allow_collaboration: bool = False,
    stop_revision: Optional[bytes] = None,
) -> PublishResult:
    """Publish a set of changes.

    Args:
      ws: Workspace to push from
      mode: Mode to use ('push', 'push-derived', 'propose')
      name: Branch name to push
      get_proposal_description: Function to retrieve proposal description
      get_proposal_commit_message: Function to retrieve proposal commit message
      dry_run: Whether to dry run
      hoster: Hoster, if known
      allow_create_proposal: Whether to allow creating proposals
      labels: Labels to set for any merge proposals
      overwrite_existing: Whether to overwrite existing (but unrelated) branch
      existing_proposal: Existing proposal to update
      reviewers: List of reviewers for merge proposal
      tags: Tags to push (None for default behaviour)
      derived_owner: Name of any derived branch
      allow_collaboration: Whether to allow target branch owners to modify
        source branch.
    """
    if mode not in SUPPORTED_MODES:
        raise ValueError("invalid mode %r" % mode)

    if stop_revision is None:
        stop_revision = local_branch.last_revision()

    if stop_revision == main_branch.last_revision():
        if existing_proposal is not None:
            logging.info("closing existing merge proposal - no new revisions")
            existing_proposal.close()
        return PublishResult(mode)

    if resume_branch and resume_branch.last_revision() == stop_revision:
        # No new revisions added on this iteration, but changes since main
        # branch. We may not have gotten round to updating/creating the
        # merge proposal last time.
        logging.info("No changes added; making sure merge proposal is up to date.")

    if hoster is None:
        hoster = get_hoster(main_branch)

    if mode == MODE_PUSH_DERIVED:
        (remote_branch, public_url) = push_derived_changes(
            local_branch,
            main_branch,
            hoster=hoster,
            name=name,
            overwrite_existing=overwrite_existing,
            tags=tags,
            owner=derived_owner,
            stop_revision=stop_revision,
        )
        return PublishResult(mode)

    if mode in (MODE_PUSH, MODE_ATTEMPT_PUSH):
        try:
            # breezy would do this check too, but we want to be *really* sure.
            with local_branch.lock_read():
                graph = local_branch.repository.get_graph()
                if not graph.is_ancestor(main_branch.last_revision(), stop_revision):
                    raise errors.DivergedBranches(main_branch, local_branch)
            push_changes(
                local_branch,
                main_branch,
                hoster=hoster,
                dry_run=dry_run,
                tags=tags,
                stop_revision=stop_revision,
            )
        except errors.PermissionDenied:
            if mode == MODE_ATTEMPT_PUSH:
                logging.info("push access denied, falling back to propose")
                mode = MODE_PROPOSE
            else:
                logging.info("permission denied during push")
                raise
        else:
            return PublishResult(mode=mode)

    assert mode == "propose"
    if not resume_branch and not allow_create_proposal:
        # TODO(jelmer): Raise an exception of some sort here?
        return PublishResult(mode)

    mp_description = get_proposal_description(
        getattr(hoster, "merge_proposal_description_format", "plain"),
        existing_proposal if resume_branch else None,
    )
    if get_proposal_commit_message is not None:
        commit_message = get_proposal_commit_message(
            existing_proposal if resume_branch else None
        )
    (proposal, is_new) = propose_changes(
        local_branch,
        main_branch,
        hoster=hoster,
        name=name,
        mp_description=mp_description,
        resume_branch=resume_branch,
        resume_proposal=existing_proposal,
        overwrite_existing=overwrite_existing,
        labels=labels,
        dry_run=dry_run,
        commit_message=commit_message,
        reviewers=reviewers,
        tags=tags,
        owner=derived_owner,
        allow_collaboration=allow_collaboration,
        stop_revision=stop_revision,
    )
    return PublishResult(mode, proposal, is_new)



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.tests import TestCaseWithTransport


from ..debian.upstream import (
    update_packaging,
)


class UpdatePackagingTests(TestCaseWithTransport):
    def test_autogen_sh(self):
        t = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.37) UNRELEASED; urgency=medium

  * Add various more aliases for salsa team names.

 -- Jelmer Vernooĳ <jelmer@debian.org>  Fri, 18 Oct 2019 17:34:35 +0000
""",
                ),
                (
                    "debian/rules",
                    """\
%:
\tdh %
""",
                ),
            ]
        )
        t.add(["debian", "debian/changelog", "debian/rules"])
        oldrev = t.commit("Initial")

        self.build_tree_contents([("autogen.sh", "#!/bin/sh\n")])
        t.add(["autogen.sh"])
        t.commit("Add autogen")

        self.addCleanup(t.lock_write().unlock)
        update_packaging(
            t,
            t.branch.repository.revision_tree(oldrev),
            committer="Jelmer Vernooĳ <jelmer@debian.org>",
        )

        self.assertFileEqual(
            """\
lintian-brush (0.37) UNRELEASED; urgency=medium

  * Add various more aliases for salsa team names.
  * Invoke autogen.sh from dh_autoreconf.

 -- Jelmer Vernooĳ <jelmer@debian.org>  Fri, 18 Oct 2019 17:34:35 +0000
""",
            "debian/changelog",
        )
        self.assertFileEqual(
            """\
%:
\tdh %

override_dh_autoreconf:
\tdh_autoreconf ./autogen.sh
""",
            "debian/rules",
        )



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatic proposal/push creation."""

import argparse
import logging
import os
import subprocess
import sys
from typing import Optional, List

import silver_platter  # noqa: F401

from breezy import osutils
from breezy import errors
from breezy.commit import PointlessCommit
from breezy.workingtree import WorkingTree
from breezy import propose as _mod_propose
from .proposal import (
    UnsupportedHoster,
    enable_tag_pushing,
    find_existing_proposed,
    get_hoster,
)
from .workspace import (
    Workspace,
)
from .publish import (
    SUPPORTED_MODES,
)
from .utils import (
    open_branch,
    BranchMissing,
    BranchUnsupported,
    BranchUnavailable,
    full_branch_url,
)


class ScriptMadeNoChanges(errors.BzrError):

    _fmt = "Script made no changes."


def script_runner(
    local_tree: WorkingTree, script: str, commit_pending: Optional[bool] = None
) -> str:
    """Run a script in a tree and commit the result.

    This ignores newly added files.

    :param local_tree: Local tree to run script in
    :param script: Script to run
    :param commit_pending: Whether to commit pending changes
        (True, False or None: only commit if there were no commits by the
         script)
    :return: Description as reported by script
    """
    last_revision = local_tree.last_revision()
    p = subprocess.Popen(
        script, cwd=local_tree.basedir, stdout=subprocess.PIPE, shell=True
    )
    (description_encoded, err) = p.communicate(b"")
    if p.returncode != 0:
        raise errors.BzrCommandError(
            "Script %s failed with error code %d" % (script, p.returncode)
        )
    new_revision = local_tree.last_revision()
    description = description_encoded.decode()
    if last_revision == new_revision and commit_pending is None:
        # Automatically commit pending changes if the script did not
        # touch the branch.
        commit_pending = True
    if commit_pending:
        try:
            new_revision = local_tree.commit(description, allow_pointless=False)
        except PointlessCommit:
            pass
    if new_revision == last_revision:
        raise ScriptMadeNoChanges()
    return description


def derived_branch_name(script: str) -> str:
    return os.path.splitext(osutils.basename(script.split(" ")[0]))[0]


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    parser = argparse.ArgumentParser()
    parser.add_argument("script", help="Path to script to run.", type=str)
    parser.add_argument("url", help="URL of branch to work on.", type=str)
    parser.add_argument(
        "--derived-owner", type=str, default=None, help="Owner for derived branches."
    )
    parser.add_argument(
        "--refresh",
        action="store_true",
        help="Refresh changes if branch already exists",
    )
    parser.add_argument(
        "--label", type=str, help="Label to attach", action="append", default=[]
    )
    parser.add_argument("--name", type=str, help="Proposed branch name", default=None)
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--mode",
        help="Mode for pushing",
        choices=SUPPORTED_MODES,
        default="propose",
        type=str,
    )
    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default="auto",
        type=str,
    )
    parser.add_argument(
        "--dry-run",
        help="Create branches but don't push or propose anything.",
        action="store_true",
        default=False,
    )
    args = parser.parse_args(argv)

    try:
        main_branch = open_branch(args.url)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.exception("%s: %s", args.url, e)
        return 1

    if args.name is None:
        name = derived_branch_name(args.script)
    else:
        name = args.name
    commit_pending = {"auto": None, "yes": True, "no": False}[args.commit_pending]

    overwrite = False

    try:
        hoster = get_hoster(main_branch)
    except UnsupportedHoster as e:
        if args.mode != "push":
            raise
        # We can't figure out what branch to resume from when there's no hoster
        # that can tell us.
        resume_branch = None
        existing_proposal = None
        logging.warn(
            "Unsupported hoster (%s), will attempt to push to %s",
            e,
            full_branch_url(main_branch),
        )
    else:
        (resume_branch, resume_overwrite, existing_proposal) = find_existing_proposed(
            main_branch, hoster, name, owner=args.derived_owner
        )
        if resume_overwrite is not None:
            overwrite = resume_overwrite
    if args.refresh:
        resume_branch = None
    with Workspace(main_branch, resume_branch=resume_branch) as ws:
        try:
            description = script_runner(ws.local_tree, args.script, commit_pending)
        except ScriptMadeNoChanges:
            logging.exception("Script did not make any changes.")
            return 1

        def get_description(description_format, existing_proposal):
            if description is not None:
                return description
            if existing_proposal is not None:
                return existing_proposal.get_description()
            raise ValueError("No description available")

        enable_tag_pushing(ws.local_tree.branch)

        try:
            publish_result = ws.publish_changes(
                args.mode,
                name,
                get_proposal_description=get_description,
                dry_run=args.dry_run,
                hoster=hoster,
                labels=args.label,
                overwrite_existing=overwrite,
                derived_owner=args.derived_owner,
                existing_proposal=existing_proposal,
            )
        except UnsupportedHoster as e:
            logging.exception(
                "No known supported hoster for %s. Run 'svp login'?",
                full_branch_url(e.branch),
            )
            return 1
        except _mod_propose.HosterLoginRequired as e:
            logging.exception(
                "Credentials for hosting site at %r missing. " "Run 'svp login'?",
                e.hoster.base_url,
            )
            return 1

        if publish_result.proposal:
            if publish_result.is_new:
                logging.info("Merge proposal created.")
            else:
                logging.info("Merge proposal updated.")
            if publish_result.proposal.url:
                logging.info("URL: %s", publish_result.proposal.url)
            logging.info("Description: %s", publish_result.proposal.get_description())

        if args.diff:
            ws.show_diff(sys.stdout.buffer)

    return None


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for uploading packages."""

import silver_platter  # noqa: F401

import datetime
from email.utils import parseaddr
import logging
import os
import subprocess
import sys
import tempfile
from typing import List

from debmutate.changelog import (
    ChangelogEditor,
    changeblock_ensure_first_line,
)
from debmutate.control import ControlEditor

from breezy import gpg
from breezy.config import extract_email_address
from breezy.errors import NoSuchTag, PermissionDenied
from breezy.commit import NullCommitReporter
from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.import_dsc import (
    DistributionBranch,
)
from breezy.plugins.debian.release import (
    release,
)
from breezy.plugins.debian.util import (
    changelog_find_previous_upload,
    dput_changes,
    find_changelog,
    debsign,
    MissingChangelogError,
)

from debian.changelog import get_maintainer

from . import (
    apt_get_source_package,
    source_package_vcs,
    split_vcs_url,
    Workspace,
    DEFAULT_BUILDER,
    select_probers,
    NoSuchPackage,
)
from ..utils import (
    open_branch,
    BranchUnavailable,
    BranchMissing,
    BranchUnsupported,
)


class LastUploadMoreRecent(Exception):
    """Last version in archive is newer than vcs version."""

    def __init__(self, archive_version, vcs_version):
        self.archive_version = archive_version
        self.vcs_version = vcs_version
        super(LastUploadMoreRecent, self).__init__(
            "last upload (%s) is more recent than vcs (%s)"
            % (archive_version, vcs_version)
        )


class NoUnuploadedChanges(Exception):
    """Indicates there are no unuploaded changes for a package."""

    def __init__(self, archive_version):
        self.archive_version = archive_version
        super(NoUnuploadedChanges, self).__init__(
            "nothing to upload, latest version is in archive: %s" % archive_version
        )


class NoUnreleasedChanges(Exception):
    """Indicates there are no unreleased changes for a package."""

    def __init__(self, version):
        self.version = version
        super(NoUnreleasedChanges, self).__init__(
            "nothing to upload, latest version in vcs is not unreleased: %s" % version
        )


class RecentCommits(Exception):
    """Indicates there are too recent commits for a package."""

    def __init__(self, commit_age, min_commit_age):
        self.commit_age = commit_age
        self.min_commit_age = min_commit_age
        super(RecentCommits, self).__init__(
            "Last commit is only %d days old (< %d)"
            % (self.commit_age, self.min_commit_age)
        )


class CommitterNotAllowed(Exception):
    """Specified committer is not allowed."""

    def __init__(self, committer, allowed_committers):
        self.committer = committer
        self.allowed_committers = allowed_committers
        super(CommitterNotAllowed, self).__init__(
            "Committer %s not in allowed committers: %r"
            % (self.committer, self.allowed_committers)
        )


class LastReleaseRevisionNotFound(Exception):
    """The revision for the last uploaded release can't be found."""

    def __init__(self, package, version):
        self.package = package
        self.version = version
        super(LastReleaseRevisionNotFound, self).__init__(
            "Unable to find revision matching version %r for %s" % (version, package)
        )


def check_revision(rev, min_commit_age, allowed_committers):
    """Check whether a revision can be included in an upload.

    Args:
      rev: revision to check
      min_commit_age: Minimum age for revisions
      allowed_committers: List of allowed committers
    Raises:
      RecentCommits: When there are commits younger than min_commit_age
    """
    # TODO(jelmer): deal with timezone
    if min_commit_age is not None:
        commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
        time_delta = datetime.datetime.now() - commit_time
        if time_delta.days < min_commit_age:
            raise RecentCommits(time_delta.days, min_commit_age)
    # TODO(jelmer): Allow tag to prevent automatic uploads
    committer_email = extract_email_address(rev.committer)
    if allowed_committers and committer_email not in allowed_committers:
        raise CommitterNotAllowed(committer_email, allowed_committers)


def find_last_release_revid(branch, version):
    db = DistributionBranch(branch, None)
    return db.revid_of_version(version)


def get_maintainer_keys(context):
    for key in context.keylist(source="/usr/share/keyrings/debian-keyring.gpg"):
        yield key.fpr
        for subkey in key.subkeys:
            yield subkey.keyid


def prepare_upload_package(
    local_tree,
    subpath,
    pkg,
    last_uploaded_version,
    builder,
    gpg_strategy=None,
    min_commit_age=None,
    allowed_committers=None,
):
    if local_tree.has_filename(os.path.join(subpath, "debian/gbp.conf")):
        subprocess.check_call(
            ["gbp", "dch", "--ignore-branch"], cwd=local_tree.abspath(".")
        )
    cl, top_level = find_changelog(local_tree, merge=False, max_blocks=None)
    if cl.version == last_uploaded_version:
        raise NoUnuploadedChanges(cl.version)
    previous_version_in_branch = changelog_find_previous_upload(cl)
    if last_uploaded_version > previous_version_in_branch:
        raise LastUploadMoreRecent(last_uploaded_version, previous_version_in_branch)

    logging.info("Checking revisions since %s" % last_uploaded_version)
    with local_tree.lock_read():
        try:
            last_release_revid = find_last_release_revid(
                local_tree.branch, last_uploaded_version
            )
        except NoSuchTag:
            raise LastReleaseRevisionNotFound(pkg, last_uploaded_version)
        graph = local_tree.branch.repository.get_graph()
        revids = list(
            graph.iter_lefthand_ancestry(
                local_tree.branch.last_revision(), [last_release_revid]
            )
        )
        if not revids:
            logging.info("No pending changes")
            return
        if gpg_strategy:
            logging.info("Verifying GPG signatures...")
            count, result, all_verifiables = gpg.bulk_verify_signatures(
                local_tree.branch.repository, revids, gpg_strategy
            )
            for revid, code, key in result:
                if code != gpg.SIGNATURE_VALID:
                    raise Exception("No valid GPG signature on %r: %d" % (revid, code))
        for revid, rev in local_tree.branch.repository.iter_revisions(revids):
            if rev is not None:
                check_revision(rev, min_commit_age, allowed_committers)

        if cl.distributions != "UNRELEASED":
            raise NoUnreleasedChanges(cl.version)
    qa_upload = False
    team_upload = False
    control_path = local_tree.abspath(os.path.join(subpath, "debian/control"))
    with ControlEditor(control_path) as e:
        maintainer = parseaddr(e.source["Maintainer"])
        if maintainer[1] == "packages@qa.debian.org":
            qa_upload = True
        # TODO(jelmer): Check whether this is a team upload
        # TODO(jelmer): determine whether this is a NMU upload
    if qa_upload or team_upload:
        changelog_path = local_tree.abspath(os.path.join(subpath, "debian/changelog"))
        with ChangelogEditor(changelog_path) as e:
            if qa_upload:
                changeblock_ensure_first_line(e[0], "QA upload.")
            elif team_upload:
                changeblock_ensure_first_line(e[0], "Team upload.")
        local_tree.commit(
            specific_files=[os.path.join(subpath, "debian/changelog")],
            message="Mention QA Upload.",
            allow_pointless=False,
            reporter=NullCommitReporter(),
        )
    tag_name = release(local_tree, subpath)
    target_dir = tempfile.mkdtemp()
    builder = builder.replace("${LAST_VERSION}", last_uploaded_version)
    target_changes = _build_helper(
        local_tree, subpath, local_tree.branch, target_dir, builder=builder
    )
    debsign(target_changes)
    return target_changes, tag_name


def select_apt_packages(package_names, maintainer):
    packages = []
    import apt_pkg

    apt_pkg.init()
    sources = apt_pkg.SourceRecords()
    while sources.step():
        if maintainer:
            fullname, email = parseaddr(sources.maintainer)
            if email not in maintainer:
                continue

        if package_names and sources.package not in package_names:
            continue

        packages.append(sources.package)

    return packages


def select_vcswatch_packages(
    packages: List[str], maintainer: List[str], autopkgtest_only: bool
):
    import psycopg2

    conn = psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net",
    )
    cursor = conn.cursor()
    args = []
    query = """\
    SELECT sources.source, vcswatch.url
    FROM vcswatch JOIN sources ON sources.source = vcswatch.source
    WHERE
     vcswatch.status IN ('COMMITS', 'NEW') AND
     sources.release = 'sid'
"""
    if autopkgtest_only:
        query += " AND sources.testsuite != '' "
    if maintainer:
        query += " AND sources.maintainer_email IN (%s)"
        args.append(tuple(maintainer))
    if packages:
        query += " AND sources.source IN (%s)"
        args.append(tuple(packages))

    cursor.execute(query, tuple(args))

    packages = []
    for package, vcs_url in cursor.fetchall():
        packages.append(package)
    return packages


def main(argv):  # noqa: C901
    import argparse

    parser = argparse.ArgumentParser(prog="upload-pending-commits")
    parser.add_argument("packages", nargs="*")
    parser.add_argument("--dry-run", action="store_true", help="Dry run changes.")
    parser.add_argument(
        "--acceptable-keys",
        help="List of acceptable GPG keys",
        action="append",
        default=[],
        type=str,
    )
    parser.add_argument(
        "--gpg-verification",
        help="Verify GPG signatures on commits",
        action="store_true",
    )
    parser.add_argument(
        "--min-commit-age",
        help="Minimum age of the last commit, in days",
        type=int,
        default=0,
    )
    parser.add_argument("--diff", action="store_true", help="Show diff.")
    parser.add_argument(
        "--builder",
        type=str,
        help="Build command",
        default=(
            DEFAULT_BUILDER + " --source --source-only-changes "
            "--debbuildopt=-v${LAST_VERSION}"
        ),
    )
    parser.add_argument(
        "--maintainer",
        type=str,
        action="append",
        help="Select all packages maintainer by specified maintainer.",
    )
    parser.add_argument(
        "--vcswatch",
        action="store_true",
        default=False,
        help="Use vcswatch to determine what packages need uploading.",
    )
    parser.add_argument(
        "--exclude", type=str, action="append", default=[], help="Ignore source package"
    )
    parser.add_argument(
        "--autopkgtest-only",
        action="store_true",
        help="Only process packages with autopkgtests.",
    )
    parser.add_argument(
        "--allowed-committer",
        type=str,
        action="append",
        help="Require that all new commits are from specified committers",
    )

    args = parser.parse_args(argv)
    ret = 0

    if not args.packages and not args.maintainer:
        (name, email) = get_maintainer()
        if email:
            logging.info("Processing packages maintained by %s", email)
            args.maintainer = [email]
        else:
            parser.print_usage()
            sys.exit(1)

    if args.vcswatch:
        packages = select_vcswatch_packages(
            args.packages, args.maintainer, args.autopkgtest_only
        )
    else:
        logging.info(
            "Use --vcswatch to only process packages for which "
            "vcswatch found pending commits."
        )
        if args.maintainer:
            packages = select_apt_packages(args.packages, args.maintainer)
        else:
            packages = args.packages

    if not packages:
        logging.info("No packages found.")
        parser.print_usage()
        sys.exit(1)

    # TODO(jelmer): Sort packages by last commit date; least recently changed
    # commits are more likely to be successful.

    if len(packages) > 1:
        logging.info("Uploading packages: %s", ", ".join(packages))

    for package in packages:
        logging.info("Processing %s", package)
        # Can't use open_packaging_branch here, since we want to use pkg_source
        # later on.
        if "/" not in package:
            try:
                pkg_source = apt_get_source_package(package)
            except NoSuchPackage:
                logging.info("%s: package not found in apt", package)
                ret = 1
                continue
            try:
                vcs_type, vcs_url = source_package_vcs(pkg_source)
            except KeyError:
                logging.info("%s: no declared vcs location, skipping", pkg_source["Package"])
                ret = 1
                continue
            source_name = pkg_source["Package"]
            if source_name in args.exclude:
                continue
            source_version = pkg_source["Version"]
            has_testsuite = "Testsuite" in pkg_source
        else:
            vcs_url = package
            vcs_type = None
            source_name = None
            source_version = None
            has_testsuite = None
        (location, branch_name, subpath) = split_vcs_url(vcs_url)
        if subpath is None:
            subpath = ""
        probers = select_probers(vcs_type)
        try:
            main_branch = open_branch(location, probers=probers, name=branch_name)
        except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
            logging.exception("%s: %s", vcs_url, e)
            ret = 1
            continue
        with Workspace(main_branch) as ws:
            if source_name is None:
                with ControlEditor(
                    ws.local_tree.abspath(os.path.join(subpath, "debian/control"))
                ) as ce:
                    source_name = ce.source["Source"]
                with ChangelogEditor(
                    ws.local_tree.abspath(os.path.join(subpath, "debian/changelog"))
                ) as cle:
                    source_version = cle[0].version
                has_testsuite = "Testsuite" in ce.source
            if source_name in args.exclude:
                continue
            if (
                args.autopkgtest_only
                and not has_testsuite
                and not ws.local_tree.has_filename(
                    os.path.join(subpath, "debian/tests/control")
                )
            ):
                logging.info(
                    "%s: Skipping, package has no autopkgtest.", source_name)
                continue
            branch_config = ws.local_tree.branch.get_config_stack()
            if args.gpg_verification:
                gpg_strategy = gpg.GPGStrategy(branch_config)
                if args.acceptable_keys:
                    acceptable_keys = args.acceptable_keys
                else:
                    acceptable_keys = list(get_maintainer_keys(gpg_strategy.context))
                gpg_strategy.set_acceptable_keys(",".join(acceptable_keys))
            else:
                gpg_strategy = None

            try:
                target_changes, tag_name = prepare_upload_package(
                    ws.local_tree,
                    subpath,
                    source_name,
                    source_version,
                    builder=args.builder,
                    gpg_strategy=gpg_strategy,
                    min_commit_age=args.min_commit_age,
                    allowed_committers=args.allowed_committer,
                )
            except CommitterNotAllowed as e:
                logging.warn(
                    "%s: committer %s not in allowed list: %r",
                    source_name,
                    e.committer,
                    e.allowed_committers,
                )
                continue
            except BuildFailedError as e:
                logging.warn("%s: package failed to build: %s", source_name, e)
                ret = 1
                continue
            except LastReleaseRevisionNotFound as e:
                logging.warn(
                    "%s: Unable to find revision matching last release "
                    "%s, skipping.",
                    source_name,
                    e.version,
                )
                ret = 1
                continue
            except LastUploadMoreRecent as e:
                logging.warn(
                    "%s: Last upload (%s) was more recent than VCS (%s)",
                    source_name,
                    e.archive_version,
                    e.vcs_version,
                )
                ret = 1
                continue
            except MissingChangelogError:
                logging.info("%s: No changelog found, skipping.", source_name)
                ret = 1
                continue
            except RecentCommits as e:
                logging.info(
                    "%s: Recent commits (%d days), skipping.",
                    source_name, e.commit_age
                )
                continue
            except NoUnuploadedChanges:
                logging.info(
                    "%s: No unuploaded changes, skipping.", source_name)
                continue
            except NoUnreleasedChanges:
                logging.info(
                    "%s: No unreleased changes, skipping.", source_name)
                continue

            tags = []
            if tag_name is not None:
                logging.info("Pushing tag %s", tag_name)
                tags.append(tag_name)
            try:
                ws.push(dry_run=args.dry_run, tags=tags)
            except PermissionDenied:
                logging.info(
                    "%s: Permission denied pushing to branch, skipping.",
                    source_name)
                ret = 1
                continue
            if not args.dry_run:
                dput_changes(target_changes)
            if args.diff:
                sys.stdout.flush()
                ws.show_diff(sys.stdout.buffer)
                sys.stdout.buffer.flush()

    return ret


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import logging
import sys
from typing import List, Set

from debian.changelog import ChangelogCreateError

import breezy

from lintian_brush import (
    available_lintian_fixers,
    run_lintian_fixers,
    DEFAULT_MINIMUM_CERTAINTY,
    SUPPORTED_CERTAINTIES,
    NotDebianPackage,
    version_string as lintian_brush_version_string,
)
from lintian_brush.config import Config

import silver_platter

from . import (
    control_files_in_root,
)
from .changer import (
    DebianChanger,
    ChangerResult,
    run_mutator,
    ChangerError,
)

__all__ = [
    "available_lintian_fixers",
    "calculate_value",
]


DEFAULT_ADDON_FIXERS = [
    "debian-changelog-line-too-long",
    "file-contains-trailing-whitespace",
    "out-of-date-standards-version",
    "package-uses-old-debhelper-compat-version",
    "public-upstream-key-not-minimal",
    "no-dh-sequencer",
]

DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY = 10
DEFAULT_VALUE_LINTIAN_BRUSH = 50
# Base these scores on the importance as set in Debian?
LINTIAN_BRUSH_TAG_VALUES = {
    "file-contains-trailing-whitespace": 0,
}
LINTIAN_BRUSH_TAG_DEFAULT_VALUE = 5


BRANCH_NAME = "lintian-fixes"


class UnknownFixer(Exception):
    """The specified fixer is unknown."""

    def __init__(self, fixer):
        self.fixer = fixer
        super(UnknownFixer, self).__init__("No such fixer: %s." % fixer)


def calculate_value(tags: Set[str]) -> int:
    if not (set(tags) - set(DEFAULT_ADDON_FIXERS)):
        value = DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY
    else:
        value = DEFAULT_VALUE_LINTIAN_BRUSH
    for tag in tags:
        value += LINTIAN_BRUSH_TAG_VALUES.get(tag, LINTIAN_BRUSH_TAG_DEFAULT_VALUE)
    return value


def parse_mp_description(description: str) -> List[str]:
    """Parse a merge proposal description.

    Args:
      description: The description to parse
    Returns:
      list of one-line descriptions of changes
    """
    existing_lines = description.splitlines()
    if len(existing_lines) == 1:
        return existing_lines
    else:
        return [
            line[2:].rstrip("\n") for line in existing_lines if line.startswith("* ")
        ]


def create_mp_description(description_format: str, lines: List[str]) -> str:
    """Create a merge proposal description.

    Args:
      lines: List of one-line descriptions of fixes
    Returns:
      A string with a merge proposal description
    """
    if len(lines) > 1:
        mp_description = ["Fix some issues reported by lintian\n"]
        for line in lines:
            line = "* %s\n" % line
            if line not in mp_description:
                mp_description.append(line)
    else:
        mp_description = [lines[0]]
    return "".join(mp_description)


def applied_entry_as_line(description_format, fixed_lintian_tags, line):
    if not fixed_lintian_tags:
        return line
    if description_format == "markdown":
        return "%s (%s)" % (
            line,
            ", ".join(
                [
                    "[%s](https://lintian.debian.org/tags/%s.html)" % (tag, tag)
                    for tag in fixed_lintian_tags
                ]
            ),
        )
    return "%s (%s)" % (line, ", ".join(fixed_lintian_tags))


def update_proposal_description(description_format, existing_proposal, applied):
    if existing_proposal:
        existing_description = existing_proposal.get_description()
        existing_lines = parse_mp_description(existing_description)
    else:
        existing_lines = []
    return create_mp_description(
        description_format,
        existing_lines
        + [
            applied_entry_as_line(description_format, r.fixed_lintian_tags, l)
            for r, l in applied
        ],
    )


def update_proposal_commit_message(existing_proposal, applied):
    existing_commit_message = getattr(
        existing_proposal, "get_commit_message", lambda: None
    )()
    if existing_commit_message and not existing_commit_message.startswith(
        "Fix lintian issues: "
    ):
        # The commit message is something we haven't set - let's leave it
        # alone.
        return
    if existing_commit_message:
        existing_applied = existing_commit_message.split(":", 1)[1]
    else:
        existing_applied = []
    return "Fix lintian issues: " + (
        ", ".join(sorted(existing_applied + [l for r, l in applied]))
    )


def has_nontrivial_changes(applied, propose_addon_only: Set[str]) -> bool:
    tags = set()
    for result, unused_summary in applied:
        tags.update(result.fixed_lintian_tags)
    # Is there enough to create a new merge proposal?
    return bool(tags - set(propose_addon_only))


def get_fixers(available_fixers, names=None, tags=None, exclude=None):
    """Get the set of fixers to try.

    Args:
      available_fixers: Dictionary mapping fixer names to objects
      names: Optional set of fixers to restrict to
      tags: Optional set of tags to restrict to
      exclude: Optional set of fixers to exclude
    Returns:
      List of fixer objects
    """
    if exclude is None:
        exclude = set()
    by_tag = {}
    by_name = {}
    for fixer in available_fixers:
        for tag in fixer.lintian_tags:
            by_tag[tag] = fixer
        by_name[fixer.name] = fixer

    # If it's unknown which fixers are relevant, just try all of them.
    if names:
        try:
            fixers = [by_name[name] for name in names]
        except KeyError as e:
            raise UnknownFixer(e.args[0])
    elif tags:
        fixers = [by_tag[tag] for tag in tags]
    else:
        fixers = list(by_name.values())

    if exclude:
        fixers = [fixer for fixer in fixers if fixer.name not in exclude]
    return fixers


class LintianBrushChanger(DebianChanger):

    name = "lintian-brush"

    def __init__(
        self,
        names=None,
        exclude=None,
        propose_addon_only=None,
        compat_release=None,
        allow_reformatting=None,
        minimum_certainty=None,
        tags=None,
        opinionated=False,
        trust_package=False,
        diligence=0,
    ):
        self.fixers = get_fixers(
            available_lintian_fixers(), names=names, tags=tags, exclude=exclude
        )
        self.propose_addon_only = propose_addon_only or []
        self.compat_release = compat_release
        self.allow_reformatting = allow_reformatting
        self.minimum_certainty = minimum_certainty
        self.opinionated = opinionated
        self.trust_package = trust_package
        self.diligence = diligence

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument(
            "--fixers", help="Fixers to run.", type=str, action="append"
        )
        parser.add_argument(
            "--exclude", help="Fixers to exclude.", type=str, action="append"
        )
        parser.add_argument(
            "--propose-addon-only",
            help="Fixers that should be considered add-on-only.",
            type=str,
            action="append",
            default=DEFAULT_ADDON_FIXERS,
        )
        parser.add_argument(
            "--compat-release",
            type=str,
            default=None,
            help="Oldest Debian release to be compatible with.",
        )
        parser.add_argument(
            "--allow-reformatting",
            default=None,
            action="store_true",
            help="Whether to allow reformatting.",
        )
        parser.add_argument(
            "--minimum-certainty",
            type=str,
            choices=SUPPORTED_CERTAINTIES,
            default=None,
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--opinionated", action="store_true", help="Make opinionated changes"
        )
        parser.add_argument("--diligence", type=int, default=10, help=argparse.SUPPRESS)
        parser.add_argument(
            "--trust-package", action="store_true", help="Trust package."
        )
        parser.add_argument("tags", nargs="*")

    @classmethod
    def from_args(cls, args):
        return cls(
            names=args.fixers,
            exclude=args.exclude,
            propose_addon_only=args.propose_addon_only,
            compat_release=args.compat_release,
            allow_reformatting=args.allow_reformatting,
            minimum_certainty=args.minimum_certainty,
            tags=args.tags,
            opinionated=args.opinionated,
            diligence=args.diligence,
            trust_package=args.trust_package,
        )

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(  # noqa: C901
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):
        base_revid = local_tree.last_revision()

        reporter.report_metadata(
            "versions",
            {
                "lintian-brush": lintian_brush_version_string,
                "silver-platter": silver_platter.version_string,
                "breezy": breezy.version_string,
            },
        )

        import distro_info

        debian_info = distro_info.DebianDistroInfo()

        compat_release = self.compat_release
        allow_reformatting = self.allow_reformatting
        minimum_certainty = None
        try:
            cfg = Config.from_workingtree(local_tree, subpath)
        except FileNotFoundError:
            pass
        else:
            compat_release = cfg.compat_release()
            if compat_release:
                compat_release = debian_info.codename(
                    compat_release, default=compat_release
                )
            allow_reformatting = cfg.allow_reformatting()
            minimum_certainty = cfg.minimum_certainty()
        if compat_release is None:
            compat_release = debian_info.stable()
        if allow_reformatting is None:
            allow_reformatting = False
        if minimum_certainty is None:
            minimum_certainty = DEFAULT_MINIMUM_CERTAINTY

        with local_tree.lock_write():
            if control_files_in_root(local_tree, subpath):
                raise ChangerError(
                    "control-files-in-root",
                    "control files live in root rather than debian/ " "(LarstIQ mode)",
                )

            try:
                overall_result = run_lintian_fixers(
                    local_tree,
                    self.fixers,
                    committer=committer,
                    update_changelog=update_changelog,
                    compat_release=compat_release,
                    allow_reformatting=allow_reformatting,
                    minimum_certainty=minimum_certainty,
                    subpath=subpath,
                    diligence=self.diligence,
                    opinionated=self.opinionated,
                    trust_package=self.trust_package,
                )
            except NotDebianPackage:
                raise ChangerError("not-debian-package", "Not a Debian package")
            except ChangelogCreateError as e:
                raise ChangerError(
                    "changelog-create-error", "Error creating changelog entry: %s" % e
                )

        applied = []
        base_applied = reporter.get_base_metadata("applied", [])
        if base_applied:
            applied.extend(base_applied)
        for result, summary in overall_result.success:
            applied.append(
                {
                    "summary": summary,
                    "description": result.description,
                    "fixed_lintian_tags": result.fixed_lintian_tags,
                    "revision_id": result.revision_id.decode("utf-8"),
                    "certainty": result.certainty,
                }
            )
        reporter.report_metadata("applied", applied)

        if overall_result.failed_fixers:
            for fixer_name, failure in overall_result.failed_fixers.items():
                logging.info("Fixer %r failed to run:", fixer_name)
                sys.stderr.write(str(failure))
        reporter.report_metadata(
            "failed",
            {name: str(e) for (name, e) in overall_result.failed_fixers.items()},
        )

        if not overall_result.success:
            raise ChangerError("nothing-to-do", "no fixers to apply")

        fixed_lintian_tags = set()
        for result, summary in overall_result.success:
            fixed_lintian_tags.update(result.fixed_lintian_tags)

        add_on_only = not has_nontrivial_changes(
            overall_result.success, self.propose_addon_only
        )

        if not reporter.get_base_metadata("add_on_only", False):
            add_on_only = False

        if not add_on_only:
            if overall_result.success:
                logging.info("only add-on fixers found")
            sufficient_for_proposal = False
            reporter.report_metadata("add_on_only", True)
        else:
            sufficient_for_proposal = True
            reporter.report_metadata("add_on_only", False)

        branches = [("main", None, base_revid, local_tree.last_revision())]

        return ChangerResult(
            description="Applied fixes for %r" % fixed_lintian_tags,
            mutator=overall_result.success,
            branches=branches,
            tags=[],
            value=calculate_value(fixed_lintian_tags),
            sufficient_for_proposal=sufficient_for_proposal,
            proposed_commit_message=update_proposal_commit_message(
                base_proposal, overall_result.success
            ),
        )

    def get_proposal_description(self, applied, description_format, existing_proposal):
        return update_proposal_description(
            description_format, existing_proposal, applied
        )

    def describe(self, applied, publish_result):
        tags = set()
        for brush_result, unused_summary in applied:
            tags.update(brush_result.fixed_lintian_tags)
        if publish_result.is_new:
            logging.info(
                "Proposed fixes %r: %s", tags, publish_result.proposal.url)
        elif tags:
            logging.info(
                "Updated proposal %s with fixes %r", publish_result.proposal.url, tags)
        else:
            logging.info(
                "No new fixes for proposal %s", publish_result.proposal.url)


if __name__ == "__main__":
    sys.exit(run_mutator(LintianBrushChanger))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from distro_info import DebianDistroInfo

import logging
import os
import re
import tempfile

from . import (
    DEFAULT_BUILDER,
)
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerResult,
)
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.util import (
    dput_changes,
    debsign,
)

from debian.changelog import format_date, get_maintainer
from debmutate.changelog import ChangelogEditor, changeblock_add_line


# See https://backports.debian.org/Contribute/


class BackportResult(object):
    def __init__(self, target_release):
        self.target_release = target_release


def backport_suffix(release):
    distro_info = DebianDistroInfo()
    version = distro_info.version(release)
    return "bpo%s" % version


def backport_distribution(release):
    distro_info = DebianDistroInfo()
    if distro_info.codename("stable") == release:
        return "%s-backports" % release
    elif distro_info.codename("oldstable") == release:
        return "%s-backports-sloppy" % release
    else:
        raise Exception("unable to determine target suite for %s" % release)


def create_bpo_version(orig_version, bpo_suffix):
    m = re.fullmatch(r"(.*)\~" + bpo_suffix + r"\+([0-9]+)", str(orig_version))
    if m:
        base = m.group(1)
        buildno = int(m.group(2)) + 1
    else:
        base = str(orig_version)
        buildno = 1
    return "%s~%s+%d" % (base, bpo_suffix, buildno)


def backport_package(local_tree, subpath, target_release, author=None):
    changes = []
    # TODO(jelmer): Check that package has a high enough popcon count,
    # and warn otherwise?
    # TODO(jelmer): Iterate Build-Depends and verify that depends are
    # satisfied by target_distribution
    # TODO(jelmer): Update Vcs-Git/Vcs-Browser header?
    target_distribution = backport_distribution(target_release)
    version_suffix = backport_suffix(target_release)
    logging.info(
        "Using target distribution %s, version suffix %s",
        target_distribution,
        version_suffix,
    )
    clp = local_tree.abspath(os.path.join(subpath, "debian/changelog"))

    if author is None:
        author = "%s <%s>" % get_maintainer()

    with ChangelogEditor(clp) as cl:
        # TODO(jelmer): If there was an existing backport, use that version
        since_version = cl[0].version
        cl.new_block(
            package=cl[0].package,
            distributions=target_distribution,
            urgency="low",
            author=author,
            date=format_date(),
            version=create_bpo_version(since_version, version_suffix),
        )
        block = cl[0]
        changeblock_add_line(
            block,
            ["Backport to %s." % target_release] + [" +" + line for line in changes],
        )

    return since_version


class BackportChanger(DebianChanger):

    name = "backport"

    def __init__(self, dry_run=False, target_release=None, builder=None):
        self.dry_run = dry_run
        self.target_release = target_release
        self.builder = builder

    @classmethod
    def setup_parser(cls, parser):
        distro_info = DebianDistroInfo()
        parser.add_argument(
            "--target-release",
            type=str,
            help="Target release",
            default=distro_info.stable(),
        )
        parser.add_argument("--dry-run", action="store_true", help="Do a dry run.")
        parser.add_argument(
            "--builder",
            type=str,
            help="Build command",
            default=(
                DEFAULT_BUILDER + " --source --source-only-changes "
                "--debbuildopt=-v${LAST_VERSION}"
            ),
        )

    @classmethod
    def from_args(cls, args):
        return cls(
            target_release=args.target_release,
            dry_run=args.dry_run,
            builder=args.builder,
        )

    def suggest_branch_name(self):
        return backport_distribution(self.target_release)

    def make_changes(
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):

        base_revision = local_tree.last_revision()

        since_version = backport_package(
            local_tree, subpath, self.target_release, author=committer
        )

        with tempfile.TemporaryDirectory() as td:
            builder = self.builder.replace("${LAST_VERSION}", str(since_version))
            target_changes = _build_helper(
                local_tree, subpath, local_tree.branch, td, builder=builder
            )
            debsign(target_changes)

            if not self.dry_run:
                dput_changes(target_changes)

        branches = [("main", None, base_revision, local_tree.last_revision())]

        # TODO(jelmer): Add debian/... tag
        tags = []

        return ChangerResult(
            description=None,
            mutator=None,
            branches=branches,
            tags=tags,
            proposed_commit_message="Backport to %s." % self.target_release,
            sufficient_for_proposal=True,
        )

    def get_proposal_description(self, applied, description_format, existing_proposal):
        return "Backport to %s." % self.target_release

    def describe(self, result, publish_result):
        if publish_result.is_new:
            logging.info(
                "Proposed backportg to %s: %s",
                result.target_release,
                publish_result.proposal.url,
            )
        else:
            logging.info("No changes for package %s", result.package_name)


if __name__ == "__main__":
    import sys

    sys.exit(run_mutator(BackportChanger))



#!/usr/bin/python
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
from typing import Optional, Callable, List, Union, Dict, BinaryIO, Any, Tuple

from breezy.branch import Branch
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.diff import show_diff_trees
from breezy.errors import (
    DivergedBranches,
    NotBranchError,
    NoColocatedBranchSupport,
)
from breezy.propose import (
    get_hoster,
    Hoster,
    MergeProposal,
    UnsupportedHoster,
)

from breezy.transport.local import LocalTransport

from .publish import (
    merge_conflicts,
    propose_changes,
    push_changes,
    push_derived_changes,
    publish_changes as _publish_changes,
    PublishResult,
)
from .utils import (
    create_temp_sprout,
    full_branch_url,
)


__all__ = [
    "Workspace",
]


class Workspace(object):
    """Workspace for creating changes to a branch.

    main_branch: The upstream branch
    resume_branch: Optional in-progress branch that we previously made changes
        on, and should ideally continue from.
    cached_branch: Branch to copy revisions from, if possible.
    local_tree: The tree the user can work in
    """

    _destroy: Optional[Callable[[], None]]
    local_tree: WorkingTree
    main_branch_revid: Optional[bytes]

    def __init__(
        self,
        main_branch: Branch,
        resume_branch: Optional[Branch] = None,
        cached_branch: Optional[Branch] = None,
        additional_colocated_branches: Optional[List[str]] = None,
        dir: Optional[str] = None,
        path: Optional[str] = None,
    ) -> None:
        self.main_branch = main_branch
        self.main_branch_revid = None
        self.cached_branch = cached_branch
        self.resume_branch = resume_branch
        self.additional_colocated_branches = additional_colocated_branches or []
        self._destroy = None
        self._dir = dir
        self._path = path

    def __str__(self):
        if self._path is None:
            return "Workspace for %s" % full_branch_url(self.main_branch)
        else:
            return "Workspace for %s at %s" % (
                full_branch_url(self.main_branch),
                self._path,
            )

    def __repr__(self):
        return (
            "%s(%r, resume_branch=%r, cached_branch=%r, "
            "additional_colocated_branches=%r, dir=%r, path=%r)"
            % (
                type(self).__name__,
                self.main_branch,
                self.resume_branch,
                self.cached_branch,
                self.additional_colocated_branches,
                self._dir,
                self._path,
            )
        )

    def __enter__(self) -> Any:
        self.local_tree, self._destroy = create_temp_sprout(
            self.cached_branch or self.resume_branch or self.main_branch,
            self.additional_colocated_branches,
            dir=self._dir,
            path=self._path,
        )
        self.main_branch_revid = self.main_branch.last_revision()
        self.refreshed = False
        with self.local_tree.branch.lock_write():
            if self.cached_branch:
                self.local_tree.pull(
                    self.resume_branch or self.main_branch, overwrite=True
                )
            if self.resume_branch:
                try:
                    self.local_tree.pull(self.main_branch, overwrite=False)
                except DivergedBranches:
                    pass
                for branch_name in self.additional_colocated_branches:
                    try:
                        remote_colo_branch = self.main_branch.controldir.open_branch(
                            name=branch_name
                        )
                    except (NotBranchError, NoColocatedBranchSupport):
                        continue
                    self.local_tree.branch.controldir.push_branch(
                        name=branch_name, source=remote_colo_branch, overwrite=True
                    )
                if merge_conflicts(self.main_branch, self.local_tree.branch):
                    logging.info("restarting branch")
                    self.local_tree.update(revision=self.main_branch_revid)
                    self.local_tree.branch.generate_revision_history(
                        self.main_branch_revid
                    )
                    self.resume_branch = None
                    self.refreshed = True
            self.orig_revid = self.local_tree.last_revision()
        return self

    def defer_destroy(self) -> Optional[Callable[[], None]]:
        ret = self._destroy
        self._destroy = None
        return ret

    def changes_since_main(self) -> bool:
        return self.local_tree.branch.last_revision() != self.main_branch_revid

    def changes_since_resume(self) -> bool:
        return self.orig_revid != self.local_tree.branch.last_revision()

    def push(
        self,
        hoster: Optional[Hoster] = None,
        dry_run: bool = False,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        stop_revision: Optional[bytes] = None,
    ) -> None:
        if hoster is None:
            try:
                hoster = get_hoster(self.main_branch)
            except UnsupportedHoster:
                if isinstance(self.main_branch.control_transport, LocalTransport):
                    hoster = None
                else:
                    raise
        return push_changes(
            self.local_tree.branch,
            self.main_branch,
            hoster=hoster,
            additional_colocated_branches=self.additional_colocated_branches,
            dry_run=dry_run,
            tags=tags,
            stop_revision=stop_revision,
        )

    def propose(
        self,
        name: str,
        description: str,
        hoster: Optional[Hoster] = None,
        existing_proposal: Optional[MergeProposal] = None,
        overwrite_existing: Optional[bool] = None,
        labels: Optional[List[str]] = None,
        dry_run: bool = False,
        commit_message: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        owner: Optional[str] = None,
        allow_collaboration: bool = False,
        stop_revision: Optional[bytes] = None,
    ) -> MergeProposal:
        if hoster is None:
            hoster = get_hoster(self.main_branch)
        return propose_changes(
            self.local_tree.branch,
            self.main_branch,
            hoster=hoster,
            name=name,
            mp_description=description,
            resume_branch=self.resume_branch,
            resume_proposal=existing_proposal,
            overwrite_existing=(overwrite_existing or False),
            labels=labels,
            dry_run=dry_run,
            commit_message=commit_message,
            reviewers=reviewers,
            owner=owner,
            additional_colocated_branches=self.additional_colocated_branches,
            tags=tags,
            allow_collaboration=allow_collaboration,
            stop_revision=stop_revision,
        )

    def push_derived(
        self,
        name: str,
        hoster: Optional[Hoster] = None,
        overwrite_existing: Optional[bool] = False,
        owner: Optional[str] = None,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        stop_revision: Optional[bytes] = None,
    ) -> Tuple[Branch, str]:
        """Push a derived branch.

        Args:
          name: Branch name
          hoster: Optional hoster to use
          overwrite_existing: Whether to overwrite an existing branch
          tags: Tags list to push
          owner: Owner name
        Returns:
          tuple with remote_branch and public_branch_url
        """
        if hoster is None:
            hoster = get_hoster(self.main_branch)
        return push_derived_changes(
            self.local_tree.branch,
            self.main_branch,
            hoster,
            name,
            overwrite_existing=overwrite_existing,
            owner=owner,
            tags=tags,
            stop_revision=stop_revision,
        )

    def publish_changes(self, *args, **kwargs) -> PublishResult:
        """Publish a set of changes."""
        return _publish_changes(
            self.local_tree.branch,
            self.main_branch,
            self.resume_branch,
            *args,
            **kwargs
        )

    def orig_tree(self) -> Tree:
        return self.local_tree.branch.repository.revision_tree(self.orig_revid)

    def show_diff(
        self, outf: BinaryIO, old_label: str = "old/", new_label: str = "new/"
    ) -> None:
        orig_tree = self.orig_tree()
        show_diff_trees(
            orig_tree,
            self.local_tree.basis_tree(),
            outf,
            old_label=old_label,
            new_label=new_label,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._destroy:
            self._destroy()
            self._destroy = None
        return False



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import contextlib
import json
import logging
import os
import subprocess
import tempfile

from urllib.request import urlopen

from debian.changelog import Changelog

from . import NoAptSources
from .changer import (
    run_mutator,
    DebianChanger,
    ChangerResult,
    ChangerError,
)
from breezy.plugins.debian.upstream import PackageVersionNotPresent


BRANCH_NAME = "missing-commits"


class AptSourceError(Exception):
    """An error occured while running 'apt source'."""

    def __init__(self, reason):
        self.reason = reason


def select_vcswatch_packages():
    import psycopg2

    conn = psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net",
    )
    cursor = conn.cursor()
    args = []
    query = """\
    SELECT sources.source, vcswatch.url
    FROM vcswatch JOIN sources ON sources.source = vcswatch.source
    WHERE
     vcswatch.status IN ('OLD', 'UNREL') AND
     sources.release = 'sid'
"""
    cursor.execute(query, tuple(args))
    packages = []
    for package, vcs_url in cursor.fetchall():
        packages.append(package)
    return packages


def download_snapshot(package, version, output_dir, no_preparation=False):
    logging.info("Downloading %s %s", package, version)
    srcfiles_url = (
        "https://snapshot.debian.org/mr/package/%s/%s/"
        "srcfiles?fileinfo=1" % (package, version)
    )
    files = {}
    for hsh, entries in json.load(urlopen(srcfiles_url))["fileinfo"].items():
        for entry in entries:
            files[entry["name"]] = hsh
    for filename, hsh in files.items():
        local_path = os.path.join(output_dir, os.path.basename(filename))
        with open(local_path, "wb") as f:
            url = "https://snapshot.debian.org/file/%s" % hsh
            with urlopen(url) as g:
                f.write(g.read())
    args = []
    if no_preparation:
        args.append("--no-preparation")
    subprocess.check_call(
        ["dpkg-source"] + args + ["-x", "%s_%s.dsc" % (package, version)],
        cwd=output_dir,
    )


class NoMissingVersions(Exception):
    def __init__(self, vcs_version, archive_version):
        self.vcs_version = vcs_version
        self.archive_version = archive_version
        super(NoMissingVersions, self).__init__(
            "No missing versions after all. Archive has %s, VCS has %s"
            % (archive_version, vcs_version)
        )


class TreeVersionNotInArchiveChangelog(Exception):
    def __init__(self, tree_version):
        self.tree_version = tree_version
        super(TreeVersionNotInArchiveChangelog, self).__init__(
            "tree version %s does not appear in archive changelog" % tree_version
        )


class TreeUpstreamVersionMissing(Exception):
    def __init__(self, upstream_version):
        self.upstream_version = upstream_version
        super(TreeUpstreamVersionMissing, self).__init__(
            "unable to find upstream version %r" % upstream_version
        )


def retrieve_source(package_name, target):
    try:
        subprocess.run(
            ["apt", "source", package_name],
            cwd=target,
            check=True,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as e:
        stderr = e.stderr.splitlines()
        if stderr[-1] == (
            b"E: You must put some 'source' URIs in your " b"sources.list"
        ):
            raise NoAptSources()
        CS = b"\x1b[1;31mE: \x1b[0m"
        CE = b"\x1b[0m"
        if stderr[-1] == (
            CS + b"You must put some 'deb-src' URIs in your sources.list" + CE
        ):
            raise NoAptSources()
        if stderr[-1].startswith(b"E: "):
            raise AptSourceError(stderr[-1][3:].decode())
        if stderr[-1].startswith(CS):
            raise AptSourceError(stderr[-1][len(CS) : -len(CE)])
        raise AptSourceError(
            [line.decode("utf-8", "surrogateescape") for line in stderr]
        )


def import_uncommitted(tree, subpath):
    from breezy.plugins.debian.import_dsc import (
        DistributionBranch,
        DistributionBranchSet,
    )

    cl_path = os.path.join(subpath, "debian/changelog")
    with tree.get_file(cl_path) as f:
        tree_cl = Changelog(f)
        package_name = tree_cl.package
    with contextlib.ExitStack() as es:
        archive_source = es.enter_context(tempfile.TemporaryDirectory())
        try:
            retrieve_source(package_name, archive_source)
        except AptSourceError as e:
            if isinstance(e.reason, list):
                reason = e.reason[-1]
            else:
                reason = e.reason
            raise ChangerError("apt-source-error", reason)
        except NoAptSources:
            raise ChangerError(
                "no-apt-sources", "No sources configured in /etc/apt/sources.list"
            )
        [subdir] = [e.path for e in os.scandir(archive_source) if e.is_dir()]
        with open(os.path.join(subdir, "debian", "changelog"), "r") as f:
            archive_cl = Changelog(f)
        missing_versions = []
        for block in archive_cl:
            if block.version == tree_cl.version:
                break
            missing_versions.append(block.version)
        else:
            raise TreeVersionNotInArchiveChangelog(tree_cl.version)
        if len(missing_versions) == 0:
            raise NoMissingVersions(tree_cl.version, archive_cl.version)
        logging.info("Missing versions: %s", ", ".join(map(str, missing_versions)))
        ret = []
        dbs = DistributionBranchSet()
        db = DistributionBranch(tree.branch, tree.branch, tree=tree)
        dbs.add_branch(db)
        if tree_cl.version.debian_revision:
            logging.info("Extracting upstream version %s.", tree_cl.version.upstream_version)
            upstream_dir = es.enter_context(tempfile.TemporaryDirectory())
            try:
                upstream_tips = db.pristine_upstream_source.version_as_revisions(
                    tree_cl.package, tree_cl.version.upstream_version
                )
            except PackageVersionNotPresent:
                raise TreeUpstreamVersionMissing(tree_cl.version.upstream_version)
            db.extract_upstream_tree(upstream_tips, upstream_dir)
        no_preparation = not tree.has_filename(".pc/applied-patches")
        version_path = {}
        for version in missing_versions:
            output_dir = es.enter_context(tempfile.TemporaryDirectory())
            download_snapshot(
                package_name, version, output_dir, no_preparation=no_preparation
            )
            version_path[version] = output_dir
        for version in reversed(missing_versions):
            logging.info("Importing %s", version)
            dsc_path = os.path.join(
                version_path[version], "%s_%s.dsc" % (package_name, version)
            )
            tag_name = db.import_package(dsc_path)
            revision = db.version_as_revisions(version)
            ret.append((tag_name, version, revision))
    return ret


class UncommittedChanger(DebianChanger):

    name = "import-upload"

    @classmethod
    def setup_parser(cls, parser):
        pass

    @classmethod
    def from_args(cls, args):
        return cls()

    def suggest_branch_name(self):
        return BRANCH_NAME

    def make_changes(
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):
        base_revid = local_tree.last_revision()
        try:
            ret = import_uncommitted(local_tree, subpath)
        except TreeUpstreamVersionMissing as e:
            raise ChangerError("tree-upstream-version-missing", str(e))
        except TreeVersionNotInArchiveChangelog as e:
            raise ChangerError("tree-version-not-in-archive-changelog", str(e))
        except NoMissingVersions as e:
            raise ChangerError("nothing-to-do", str(e))
        tags = [(None, tag_name, revid) for (tag_name, version, revid) in ret]
        # TODO(jelmer): Include upstream tags
        proposed_commit_message = "Import missing uploads: %s." % (
            ", ".join([str(v) for t, v in ret])
        )
        reporter.report_metadata(
            "tags", [(tag_name, str(version)) for (tag_name, version, revid) in ret]
        )

        branches = [("main", None, base_revid, local_tree.last_revision())]

        # TODO(jelmer): Include branches for upstream/pristine-tar

        return ChangerResult(
            description="Import archive changes missing from the VCS.",
            branches=branches,
            mutator=ret,
            tags=tags,
            sufficient_for_proposal=True,
            proposed_commit_message=proposed_commit_message,
        )

    def get_proposal_description(self, applied, description_format, existing_proposal):
        return "Import missing uploads: %s." % (", ".join([str(v) for t, v in applied]))

    def describe(self, applied, publish_result):
        if publish_result.is_new:
            logging.info(
                "Proposed import of versions %s: %s",
                ", ".join([str(v) for t, v in applied]),
                publish_result.proposal.url,
            )
        elif applied:
            logging.info(
                "Updated proposal %s with versions %s.",
                publish_result.proposal.url,
                ", ".join([str(v) for t, v in applied]),
            )
        else:
            logging.info(
                "No new versions imported for proposal %s", publish_result.proposal.url
            )

    @classmethod
    def describe_command(cls, command):
        return "Import archive changes missing from VCS"


if __name__ == "__main__":
    import sys

    sys.exit(run_mutator(UncommittedChanger))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = ["iter_conflicted"]

import argparse
from functools import partial
import logging
import pkg_resources
import sys
from typing import Any, List, Optional, Dict, Iterable, Tuple, Type

from breezy import version_info as breezy_version_info
from breezy.branch import Branch
from breezy.propose import Hoster, MergeProposal
from breezy.transport import Transport
from breezy.workingtree import WorkingTree

from . import (
    open_packaging_branch,
    guess_update_changelog,
    NoSuchPackage,
    NoAptSources,
    DEFAULT_BUILDER,
)
from ..proposal import (
    HosterLoginRequired,
    UnsupportedHoster,
    NoSuchProject,
    enable_tag_pushing,
    find_existing_proposed,
    get_hoster,
    iter_conflicted,
)

from ..publish import (
    PublishResult,
    SUPPORTED_MODES,
)
from ..utils import (
    BranchMissing,
    BranchUnavailable,
    BranchUnsupported,
    run_pre_check,
    run_post_check,
    PostCheckFailed,
    full_branch_url,
)


def get_package(
    package: str,
    branch_name: str,
    overwrite_unrelated: bool = False,
    refresh: bool = False,
    possible_transports: Optional[List[Transport]] = None,
    possible_hosters: Optional[List[Hoster]] = None,
    owner: Optional[str] = None,
) -> Tuple[
    str,
    Branch,
    str,
    Optional[Branch],
    Optional[Hoster],
    Optional[MergeProposal],
    Optional[bool],
]:
    main_branch, subpath = open_packaging_branch(
        package, possible_transports=possible_transports
    )

    overwrite: Optional[bool] = False

    try:
        hoster = get_hoster(main_branch, possible_hosters=possible_hosters)
    except UnsupportedHoster:
        # We can't figure out what branch to resume from when there's no
        # hoster that can tell us.
        resume_branch = None
        existing_proposal = None
        hoster = None
    else:
        (resume_branch, overwrite, existing_proposal) = find_existing_proposed(
            main_branch,
            hoster,
            branch_name,
            owner=owner,
            overwrite_unrelated=overwrite_unrelated,
        )
    if refresh:
        overwrite = True
        resume_branch = None

    return (
        package,
        main_branch,
        subpath,
        resume_branch,
        hoster,
        existing_proposal,
        overwrite,
    )


def iter_packages(
    packages: Iterable[str],
    branch_name: str,
    overwrite_unrelated: bool = False,
    refresh: bool = False,
    derived_owner: Optional[str] = None,
):
    """Iterate over relevant branches for a set of packages.

    Args:
      packages: Iterable over package names (or packaging URLs)
      branch_name: Branch name to look for
      overwrite_unrelated: Allow overwriting unrelated changes
      refresh: Whether to refresh existing merge proposals
    Returns:
      iterator over
        (package name, main branch object, subpath, branch to resume (if any),
         hoster (None if the hoster is not supported),
         existing_proposal, whether to overwrite the branch)
    """
    possible_transports: List[Transport] = []
    possible_hosters: List[Hoster] = []

    for pkg in packages:
        logging.info("Processing: %s", pkg)

        (
            pkg,
            main_branch,
            subpath,
            resume_branch,
            hoster,
            existing_proposal,
            overwrite,
        ) = get_package(
            pkg,
            branch_name,
            overwrite_unrelated=overwrite_unrelated,
            refresh=refresh,
            possible_transports=possible_transports,
            possible_hosters=possible_hosters,
            owner=derived_owner,
        )

        yield (
            pkg,
            main_branch,
            subpath,
            resume_branch,
            hoster,
            existing_proposal,
            overwrite,
        )


class ChangerReporter(object):
    def report_context(self, context):
        raise NotImplementedError(self.report_context)

    def report_metadata(self, key, value):
        raise NotImplementedError(self.report_metadata)

    def get_base_metadata(self, key, default_value=None):
        raise NotImplementedError(self.get_base_metadata)


class ChangerError(Exception):
    def __init__(
        self, category: str, summary: str, original: Optional[Exception] = None
    ):
        self.category = category
        self.summary = summary
        self.original = original


class ChangerResult(object):
    def __init__(
        self,
        description: Optional[str],
        mutator: Any,
        branches: Optional[List[Tuple[str, str, bytes, bytes]]] = [],
        tags: Optional[Dict[str, bytes]] = None,
        value: Optional[int] = None,
        proposed_commit_message: Optional[str] = None,
        title: Optional[str] = None,
        labels: Optional[List[str]] = None,
        sufficient_for_proposal: bool = True,
    ):
        self.description = description
        self.mutator = mutator
        self.branches = branches or []
        self.tags = tags or {}
        self.value = value
        self.proposed_commit_message = proposed_commit_message
        self.title = title
        self.labels = labels
        self.sufficient_for_proposal = sufficient_for_proposal

    def show_diff(
        self,
        repository,
        outf,
        role="main",
        old_label: str = "old/",
        new_label: str = "new/",
    ) -> None:
        from breezy.diff import show_diff_trees

        for (brole, name, base_revision, revision) in self.branches:
            if role == brole:
                break
        else:
            raise KeyError
        old_tree = repository.revision_tree(base_revision)
        new_tree = repository.revision_tree(revision)
        show_diff_trees(
            old_tree, new_tree, outf, old_label=old_label, new_label=new_label
        )


def setup_parser_common(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--dry-run",
        help="Create branches but don't push or propose anything.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--build-verify",
        help="Build package to verify it.",
        dest="build_verify",
        action="store_true",
    )
    parser.add_argument(
        "--pre-check",
        help="Command to run to check whether to process package.",
        type=str,
    )
    parser.add_argument(
        "--post-check", help="Command to run to check package before pushing.", type=str
    )
    parser.add_argument(
        "--builder",
        default=DEFAULT_BUILDER,
        type=str,
        help="Build command to use when verifying build.",
    )
    parser.add_argument(
        "--refresh",
        help="Discard old branch and apply fixers from scratch.",
        action="store_true",
    )
    parser.add_argument("--committer", help="Committer identity", type=str)
    parser.add_argument(
        "--mode",
        help="Mode for pushing",
        choices=SUPPORTED_MODES,
        default="propose",
        type=str,
    )
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--diff", action="store_true", help="Output diff of created merge proposal."
    )
    parser.add_argument(
        "--build-target-dir",
        type=str,
        help=(
            "Store built Debian files in specified directory " "(with --build-verify)"
        ),
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing branches."
    )
    parser.add_argument("--name", type=str, help="Proposed branch name", default=None)
    parser.add_argument(
        "--derived-owner", type=str, default=None, help="Owner for derived branches."
    )
    parser.add_argument(
        "--label", type=str, help="Label to attach", action="append", default=[]
    )


class DebianChanger(object):
    """A class which can make and explain changes to a Debian package in VCS."""

    name: str

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser) -> None:
        raise NotImplementedError(cls.setup_parser)

    @classmethod
    def from_args(cls, args: List[str]) -> "DebianChanger":
        raise NotImplementedError(cls.from_args)

    def suggest_branch_name(self) -> str:
        raise NotImplementedError(self.suggest_branch_name)

    def make_changes(
        self,
        local_tree: WorkingTree,
        subpath: str,
        update_changelog: bool,
        reporter: ChangerReporter,
        committer: Optional[str],
        base_proposal: Optional[MergeProposal] = None,
    ) -> ChangerResult:
        raise NotImplementedError(self.make_changes)

    def get_proposal_description(
        self, applied: Any, description_format: str, existing_proposal: MergeProposal
    ) -> str:
        raise NotImplementedError(self.get_proposal_description)

    def describe(self, applied: Any, publish_result: PublishResult) -> None:
        raise NotImplementedError(self.describe)

    @classmethod
    def describe_command(cls, command):
        return cls.name


class DummyChangerReporter(ChangerReporter):
    def report_context(self, context):
        pass

    def report_metadata(self, key, value):
        pass

    def get_base_metadata(self, key, default_value=None):
        return None


def _run_single_changer(  # noqa: C901
    changer: DebianChanger,
    pkg: str,
    main_branch: Branch,
    subpath: str,
    resume_branch: Optional[Branch],
    hoster: Optional[Hoster],
    existing_proposal: Optional[MergeProposal],
    overwrite: Optional[bool],
    mode: str,
    branch_name: str,
    diff: bool = False,
    committer: Optional[str] = None,
    build_verify: bool = False,
    pre_check: Optional[str] = None,
    post_check: Optional[str] = None,
    builder: str = DEFAULT_BUILDER,
    dry_run: bool = False,
    update_changelog: Optional[bool] = None,
    label: Optional[List[str]] = None,
    derived_owner: Optional[str] = None,
    build_target_dir: Optional[str] = None,
) -> Optional[bool]:
    from breezy import errors
    from . import (
        BuildFailedError,
        MissingUpstreamTarball,
        Workspace,
    )

    if hoster is None and mode == "attempt-push":
        logging.warn(
            "Unsupported hoster; will attempt to push to %s",
            full_branch_url(main_branch),
        )
        mode = "push"
    with Workspace(
        main_branch, resume_branch=resume_branch
    ) as ws, ws.local_tree.lock_write():
        if ws.refreshed:
            overwrite = True
        run_pre_check(ws.local_tree, pre_check)
        if update_changelog is None:
            dch_guess = guess_update_changelog(ws.local_tree)
            if dch_guess:
                logging.info(dch_guess[1])
                update_changelog = dch_guess[0]
            else:
                # Assume yes.
                update_changelog = True
        try:
            changer_result = changer.make_changes(
                ws.local_tree,
                subpath=subpath,
                update_changelog=update_changelog,
                committer=committer,
                reporter=DummyChangerReporter(),
            )
        except ChangerError as e:
            logging.exception(e.summary)
            return False

        if not ws.changes_since_main():
            if existing_proposal:
                logging.info("%s: nothing left to do. Closing proposal.", pkg)
                existing_proposal.close()
            else:
                logging.info("%s: nothing to do", pkg)
            return None

        try:
            run_post_check(ws.local_tree, post_check, ws.orig_revid)
        except PostCheckFailed as e:
            logging.info("%s: %s", pkg, e)
            return False
        if build_verify:
            try:
                ws.build(builder=builder, result_dir=build_target_dir)
            except BuildFailedError:
                logging.info("%s: build failed", pkg)
                return False
            except MissingUpstreamTarball:
                logging.info("%s: unable to find upstream source", pkg)
                return False

        enable_tag_pushing(ws.local_tree.branch)

        kwargs: Dict[str, Any] = {}
        if breezy_version_info >= (3, 1):
            kwargs["tags"] = changer_result.tags

        try:
            publish_result = ws.publish_changes(
                mode,
                branch_name,
                get_proposal_description=partial(
                    changer.get_proposal_description, changer_result.mutator
                ),
                get_proposal_commit_message=(
                    lambda oldmp: changer_result.proposed_commit_message
                ),
                dry_run=dry_run,
                hoster=hoster,
                allow_create_proposal=changer_result.sufficient_for_proposal,
                overwrite_existing=overwrite,
                existing_proposal=existing_proposal,
                derived_owner=derived_owner,
                labels=label,
                **kwargs
            )
        except UnsupportedHoster as e:
            logging.exception(
                "%s: No known supported hoster for %s. Run 'svp login'?",
                pkg,
                full_branch_url(e.branch),
            )
            return False
        except NoSuchProject as e:
            logging.info("%s: project %s was not found", pkg, e.project)
            return False
        except errors.PermissionDenied as e:
            logging.info("%s: %s", pkg, e)
            return False
        except errors.DivergedBranches:
            logging.info("%s: a branch exists. Use --overwrite to discard it.", pkg)
            return False
        except HosterLoginRequired as e:
            logging.exception(
                "Credentials for hosting site at %r missing. " "Run 'svp login'?",
                e.hoster.base_url,
            )
            return False

        if publish_result.proposal:
            changer.describe(changer_result.mutator, publish_result)
        if diff:
            for entry in changer_result.branches:
                role = entry[0]
                if len(changer_result.branches) > 1:
                    sys.stdout.write("%s\n" % role)
                    sys.stdout.write(("-" * len(role)) + "\n")
                sys.stdout.flush()
                changer_result.show_diff(
                    ws.local_tree.branch.repository, sys.stdout.buffer, role=role
                )
                if len(changer_result.branches) > 1:
                    sys.stdout.write("\n")

        return True


def run_single_changer(changer: DebianChanger, args: argparse.Namespace) -> int:
    import silver_platter  # noqa: F401

    if args.name:
        branch_name = args.name
    else:
        branch_name = changer.suggest_branch_name()

    try:
        (
            pkg,
            main_branch,
            subpath,
            resume_branch,
            hoster,
            existing_proposal,
            overwrite,
        ) = get_package(
            args.package,
            branch_name,
            overwrite_unrelated=args.overwrite,
            refresh=args.refresh,
            owner=args.derived_owner,
        )
    except NoSuchPackage:
        logging.info("%s: no such package", args.package)
        return 1
    except NoSuchProject as e:
        logging.info("%s: unable to find project: %s", args.package, e.project)
        return 1
    except (BranchMissing, BranchUnavailable, BranchUnsupported) as e:
        logging.info("%s: ignoring: %s", args.package, e)
        return 1
    except NoAptSources:
        logging.info(
            "%s: no apt sources configured, unable to get package metadata.",
            args.package,
        )
        return 1

    if (
        _run_single_changer(
            changer,
            pkg,
            main_branch,
            subpath,
            resume_branch,
            hoster,
            existing_proposal,
            overwrite,
            args.mode,
            branch_name,
            diff=args.diff,
            committer=args.committer,
            build_verify=args.build_verify,
            pre_check=args.pre_check,
            builder=args.builder,
            post_check=args.post_check,
            dry_run=args.dry_run,
            update_changelog=args.update_changelog,
            label=args.label,
            derived_owner=args.derived_owner,
            build_target_dir=args.build_target_dir,
        )
        is False
    ):
        return 1
    else:
        return 0


BUILTIN_ENTRYPOINTS = [
    pkg_resources.EntryPoint(
        "run", "silver_platter.debian.run", attrs=("ScriptChanger",)
    ),
    pkg_resources.EntryPoint(
        "lintian-brush", "silver_platter.debian.lintian", attrs=("LintianBrushChanger",)
    ),
    pkg_resources.EntryPoint(
        "tidy", "silver_platter.debian.tidy", attrs=("TidyChanger",)
    ),
    pkg_resources.EntryPoint(
        "new-upstream", "silver_platter.debian.upstream", attrs=("NewUpstreamChanger",)
    ),
    pkg_resources.EntryPoint("cme-fix", "silver_platter.debian.cme", attrs=("CMEFixChanger",)),
    pkg_resources.EntryPoint(
        "apply-multiarch-hints",
        "silver_platter.debian.multiarch",
        attrs=("MultiArchHintsChanger",),
    ),
    pkg_resources.EntryPoint(
        "rules-requires-root",
        "silver_platter.debian.rrr",
        attrs=("RulesRequiresRootChanger",),
    ),
    pkg_resources.EntryPoint(
        "orphan", "silver_platter.debian.orphan", attrs=("OrphanChanger",)
    ),
    pkg_resources.EntryPoint(
        "import-upload",
        "silver_platter.debian.uncommitted",
        attrs=("UncommittedChanger",),
    ),
    pkg_resources.EntryPoint(
        "scrub-obsolete",
        "silver_platter.debian.scrub_obsolete",
        attrs=("ScrubObsoleteChanger",),
    ),
    pkg_resources.EntryPoint(
        "debianize", "silver_platter.debian.debianize", attrs=("DebianizeChanger",)
    ),
]


def changer_subcommands() -> List[str]:
    endpoints = pkg_resources.iter_entry_points("silver_platter.debian.changer")
    ret = []
    for ep in BUILTIN_ENTRYPOINTS + list(endpoints):
        ret.append(ep.name)
    return ret


def changer_subcommand(name: str) -> Type[DebianChanger]:
    for ep in BUILTIN_ENTRYPOINTS:
        if ep.name == name:
            return ep.resolve()
    endpoints = pkg_resources.iter_entry_points("silver_platter.debian.changer", name)
    for ep in endpoints:
        return ep.load()
    raise KeyError(name)


def run_mutator(changer_cls, argv=None):
    import json
    import argparse
    import os
    from breezy.workingtree import WorkingTree

    parser = argparse.ArgumentParser()
    changer_cls.setup_parser(parser)
    args = parser.parse_args(argv)
    wt, subpath = WorkingTree.open_containing(".")
    changer = changer_cls.from_args(args)
    try:
        update_changelog_str = (os.environ["UPDATE_CHANGELOG"],)
    except KeyError:
        update_changelog = None
    else:
        if update_changelog_str == "leave_changelog":
            update_changelog = False
        elif update_changelog_str == "update_changelog":
            update_changelog = True
        else:
            # TODO(jelmer): Warn
            update_changelog = None

    try:
        base_metadata_path = os.environ["BASE_METADATA"]
    except KeyError:
        existing_proposal = None
    else:
        with open(base_metadata_path, "r") as f:
            base_metadata = json.load(f)

        class PreviousProposal(MergeProposal):
            def __init__(self, metadata):
                self.metadata = metadata

            def get_description(self):
                return self.metadata.get("description")

            def get_commit_message(self):
                return self.metadata.get("commit-message")

        existing_proposal = PreviousProposal(base_metadata["merge-proposal"])

    mutator_metadata = {}
    try:
        result = changer.make_changes(
            wt,
            subpath,
            update_changelog=update_changelog,
            committer=os.environ.get("COMMITTER"),
            base_proposal=existing_proposal,
            reporter=DummyChangerReporter(),
        )
    except ChangerError as e:
        result_json = {
            "result-code": e.category,
            "description": e.summary,
        }
    else:
        result_json = {
            "result-code": None,
            "description": result.description,
            "suggested-branch-name": changer.suggest_branch_name(),
            "tags": result.tags,
            "branches": result.branches,
            "value": result.value,
            "mutator": mutator_metadata,
            "merge-proposal": {
                "sufficient": changer.sufficient_for_proposal,
                "commit-message": result.proposed_commit_message,
                "title": result.title,
                "labels": result.labels,
                "description-plain": changer.get_proposal_description(
                    result.mutator, "plain", existing_proposal
                ),
                "description-markdown": changer.get_proposal_description(
                    result.mutator, "markdown", existing_proposal
                ),
            },
        }
    json.dump(result_json, sys.stdout, indent=4)
    return 0



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for updating with a script."""

import logging

from .changer import (
    ChangerError,
    ChangerResult,
    DebianChanger,
)
from ..run import (
    ScriptMadeNoChanges,
    derived_branch_name,
    script_runner,
)


class ScriptChanger(DebianChanger):

    name = "run"

    def _init__(self, script, commit_pending=None):
        self.script = script
        self.commit_pending = commit_pending

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument("script", help="Path to script to run.", type=str)
        parser.add_argument(
            "--commit-pending",
            help="Commit pending changes after script.",
            choices=["yes", "no", "auto"],
            default="auto",
            type=str,
        )

    @classmethod
    def from_args(cls, args):
        commit_pending = {"auto": None, "yes": True, "no": False}[args.commit_pending]
        return cls(script=args.script, commit_pending=commit_pending)

    def make_changes(
        self,
        local_tree,
        subpath,
        update_changelog,
        reporter,
        committer,
        base_proposal=None,
    ):
        base_revid = local_tree.last_revision()

        try:
            description = script_runner(local_tree, self.script, self.commit_pending)
        except ScriptMadeNoChanges as e:
            raise ChangerError("nothing-to-do", "Script did not make any changes.", e)

        branches = [("main", None, base_revid, local_tree.last_revision())]

        tags = []

        # TODO(jelmer): Compare old and new tags/branches?

        return ChangerResult(
            description=description,
            mutator=description,
            sufficient_for_proposal=True,
            branches=branches,
            tags=tags,
            proposed_commit_message=None,
        )

    def get_proposal_description(
        self, description, description_format, existing_proposal
    ):
        if description is not None:
            return description
        if existing_proposal is not None:
            return existing_proposal.get_description()
        raise ValueError("No description available")

    def describe(self, description, publish_result):
        logging.info("%s", description)

    def suggest_branch_name(self):
        return derived_branch_name(self.script)
